# This code is based on:
#   http://lists.wxwidgets.org/archive/wxPython-users/msg11078.html


from __future__ import annotations

import math
import os
import time

import gltbx.fonts
import gltbx.gl as gl
import gltbx.gl_managed
import gltbx.glu as glu
import gltbx.images
import gltbx.util
import scitbx.math
from scitbx import matrix
from scitbx.array_family import flex

try:
    import wx
except ImportError:
    exit()  # To pass through the "make" step (dials.geometry_viewer), for graphics-free HPC build
import wx.glcanvas


def animation_stepper(time_move=1.0, move_factor=1, frames_per_second=100):
    time_move *= move_factor
    n_steps = max(1, int(time_move * frames_per_second + 0.5))
    time_per_frame = time_move / n_steps
    i_step = 1
    t0 = time.time()
    while True:
        f = min(1, i_step / n_steps)
        yield f
        if f == 1:
            break
        ideal_t = i_step * time_per_frame
        i_step += 1
        delta_t = time.time() - t0
        if delta_t < ideal_t:
            time.sleep(ideal_t - delta_t)
        else:
            i_step = max(i_step, int(delta_t / time_per_frame + 0.5))


def v3distsq(a, b):
    result = 0
    for x, y in zip(a, b):
        result += (x - y) ** 2
    return result


VIEWER_UPDATE_ID = wx.NewId()


class ViewerUpdateEvent(wx.PyEvent):
    def __init__(self, data=None, recenter=False):
        wx.PyEvent.__init__(self)
        self.data = data
        self.recenter = recenter
        self.SetEventType(VIEWER_UPDATE_ID)


class wxGLWindow(wx.glcanvas.GLCanvas):
    def InitGL(self):
        raise NotImplementedError()

    def DrawGL(self):
        raise NotImplementedError()

    def process_keyword_arguments(
        self,
        auto_spin_allowed=False,
        orthographic=False,
        animation_time=1,  # seconds
        background_rgb=(0, 0, 0),
        **kw,
    ):
        self.autospin_allowed = auto_spin_allowed
        self.orthographic = orthographic
        self.animation_time = animation_time
        self.background_rgb = background_rgb
        return kw

    def __init__(self, parent, *args, **kw):
        kw["attribList"] = kw.get("attribList", [])

        # Safely identify supported attributes
        if hasattr(wx.glcanvas, "WX_GL_RGBA"):
            kw["attribList"].append(wx.glcanvas.WX_GL_RGBA)  # required on Windows

        # Find the highest reasonable depth buffer this display supports - clearly
        # if none are supported, no setting will be set
        for buffer_depth in [24, 16, 8, 4]:
            if self.IsDisplaySupported([wx.glcanvas.WX_GL_DEPTH_SIZE, buffer_depth]):
                kw["attribList"].extend([wx.glcanvas.WX_GL_DEPTH_SIZE, buffer_depth])
                break

        if hasattr(wx.glcanvas, "WX_GL_DOUBLEBUFFER"):
            try:
                if self.IsDisplaySupported([wx.glcanvas.WX_GL_DOUBLEBUFFER]):
                    kw["attribList"].append(wx.glcanvas.WX_GL_DOUBLEBUFFER)
            except AttributeError:
                # IsDisplaySupported may not be present for wxPython < 2.9
                # Still attempt to enable double buffering
                kw["attribList"].append(wx.glcanvas.WX_GL_DOUBLEBUFFER)

        try:
            if hasattr(wx.glcanvas, "WX_GL_SAMPLE_BUFFERS"):
                if self.IsDisplaySupported(
                    [wx.glcanvas.WX_GL_SAMPLE_BUFFERS, gl.GL_TRUE]
                ):
                    kw["attribList"].extend(
                        [wx.glcanvas.WX_GL_SAMPLE_BUFFERS, gl.GL_TRUE]
                    )
            if hasattr(wx.glcanvas, "WX_GL_SAMPLES"):
                if self.IsDisplaySupported([wx.glcanvas.WX_GL_SAMPLES, 4]):
                    kw["attribList"].extend([wx.glcanvas.WX_GL_SAMPLES, 4])
        except AttributeError:
            # IsDisplaySupported may not be present for wxPython < 2.9
            # In that case do not attempt to use multi-/supersampling
            pass

        kw = self.process_keyword_arguments(**kw)
        self.GL_uninitialised = 1
        wx.glcanvas.GLCanvas.__init__(*((self, parent) + args), **kw)
        self.context = wx.glcanvas.GLContext(self)

        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftClick)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_MIDDLE_DOWN, self.OnMiddleClick)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightClick)
        self.Bind(wx.EVT_MOTION, self.OnMouseMotion)
        self.Bind(wx.EVT_IDLE, self.OnIdle)
        self.Bind(wx.EVT_TIMER, self.OnTimer)
        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
        self.Connect(-1, -1, VIEWER_UPDATE_ID, self.OnUpdate)

        self.spintimer_id = 100  # any number
        self.spintimer_maxfps = 30
        self.spintimer = wx.Timer(self, self.spintimer_id)
        self.spintimer.Start(1000 / self.spintimer_maxfps)

        self.w, self.h = self.GetClientSize()

        self.field_of_view_y = 10.0
        self.min_near = 1
        self.min_dist = -100
        self.min_viewport_use_fraction = 0.01
        self.slab_scale = 1.0
        self.scale_max = 0.1
        self.z_min = 4.0
        self.fog_scale_factor = 0.5
        self.flag_show_fog = False  # leave off by default
        self._settings_widget = None

        self.flag_use_lights = False

        self.rotation_center = (0, 0, 0)
        self.marked_rotation = None

        self.parent = parent
        # Current coordinates of the mouse.
        self.xmouse = 0
        self.ymouse = 0

        self.xspin = 0
        self.yspin = 0

        # Is the widget currently autospinning?
        self.autospin = False

        self.initLeft = (0, 0)
        self.was_dragged = False
        self.pick_points = None
        self.light0_position = [0, 0, 100, 0]

    def OnEraseBackground(self, event):
        pass  # Do nothing, to avoid flashing on MSW.

    def OnSize(self, event=None):
        self.w, self.h = self.GetClientSize()
        if self.IsShownOnScreen():
            if self.context is not None:
                if wx.VERSION[0] != 3:
                    self.SetCurrent(self.context)
                else:
                    self.SetCurrent()
            gl.glViewport(0, 0, self.w, self.h)

    def OnIdle(self, event):
        pass

    def OnTimer(self, event):
        if self.autospin:
            self.do_AutoSpin()

    def OnChar(self, event):
        key = event.GetKeyCode()
        if key == ord("m"):
            self.move_rotation_center_to_mcs_center()
        elif key == ord("c"):
            self.move_to_center_of_viewport(self.rotation_center)
        elif key == ord("f"):
            self.fit_into_viewport()
        elif key == ord("k"):
            self.mark_rotation()
        elif key == ord("a"):
            self.snap_back_rotation()
        elif key == ord("s"):
            self.autospin_allowed = not self.autospin_allowed
        elif key == ord("l"):
            self.flag_show_labels = not self.flag_show_labels
            self.OnRedraw()
        elif key == ord("S"):
            self.save_screen_shot()
        elif key == ord("V"):
            gltbx.util.show_versions()
        elif key == ord("O"):
            self.edit_opengl_settings()
        elif key == ord("\t"):
            callback = getattr(self, "tab_callback", None)
            if callback is None:
                print("Tab callback not available.")
            else:
                kwargs = {"shift_down": event.ShiftDown()}
                if event.ControlDown():
                    kwargs["control_down"] = True
                callback(**kwargs)
        else:
            callback = getattr(self, "process_key_stroke", None)
            if callback is None:
                print("No action for this key stroke.")
            else:
                if callback(key=key) is False:
                    print("No action for this key stroke.")
        self.autospin = False

    def OnMouseWheel(self, event):
        scale = 0.002 * event.GetWheelRotation()
        self.OnScale(scale)

    def OnLeftClick(self, event):
        self.OnRecordMouse(event)
        self.initLeft = event.GetX(), event.GetY()
        self.was_dragged = False
        self.autospin = False

    def OnLeftUp(self, event):
        if not self.was_dragged:
            self.get_pick_points(self.initLeft)
            self.process_pick_points()
            self.OnRedraw()
        else:
            self.was_dragged = False
            if not event.ShiftDown():
                self.OnAutoSpin(event)

    def OnMiddleClick(self, event):
        self.OnRecordMouse(event)

    def OnRightClick(self, event):
        self.OnRecordMouse(event)

    def OnLeftDrag(self, event):
        self.was_dragged = True
        if event.AltDown():
            self.OnTranslate(event)
        else:
            self.OnRotate(event)

    def OnMiddleDrag(self, event):
        self.OnTranslate(event)

    def OnRightDrag(self, event):
        scale = 0.02 * (event.GetY() - self.ymouse)
        self.OnScale(scale)
        self.OnRecordMouse(event)

    def OnMouseMotion(self, event):
        if not event.Dragging():
            return
        if event.LeftIsDown():
            self.OnLeftDrag(event)
        elif event.MiddleIsDown():
            self.OnMiddleDrag(event)
        elif event.RightIsDown():
            self.OnRightDrag(event)

    def setup_viewing_volume(self):
        aspect = self.w / max(1, self.h)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        near, far = self.get_clipping_distances()
        if self.orthographic:
            s = self.minimum_covering_sphere
            # c = s.center()
            c = self.rotation_center
            r = s.radius()
            rf = self.buffer_factor * r
            left = c[0] - rf
            right = c[0] + rf
            bottom = c[1] - rf
            top = c[1] + rf
            if aspect < 1:
                bottom /= aspect
                top /= aspect
            else:
                left *= aspect
                right *= aspect
            gl.glOrtho(left, right, bottom, top, near, far)
        else:
            glu.gluPerspective(self.field_of_view_y, aspect, near, far)
        self.set_lights()
        self.setup_fog()

    def get_clipping_distances(self):
        slab = self.far - self.near
        clip = (1.0 - self.slab_scale) * (slab / 2.0)
        near = self.near + clip
        far = self.far - clip
        if near > far:
            near = far - 1
        if near < self.min_near:
            near = self.min_near
        return (near, far)

    def setup_lighting(self):
        if self.flag_use_lights:
            gl.glEnable(gl.GL_LIGHTING)
            gl.glEnable(gl.GL_LIGHT0)
            gl.glEnable(gl.GL_BLEND)
            # gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
            gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ZERO)
            gl.glLightModeli(gl.GL_LIGHT_MODEL_TWO_SIDE, gl.GL_TRUE)
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_AMBIENT, [0.0, 0.0, 0.0, 1.0])
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, [1, 1, 1, 1])
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, [0.5, 0.5, 0.5, 1.0])

    def set_lights(self):
        if self.flag_use_lights:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPushMatrix()
            gl.glLoadIdentity()
            gl.glEnable(gl.GL_LIGHTING)
            gl.glEnable(gl.GL_LIGHT0)
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, self.light0_position)
            gl.glPopMatrix()

    def setup_fog(self):
        if self.flag_show_fog:
            near, far = self.get_clipping_distances()
            fog_start = near + self.fog_scale_factor * (far - near)
            fog_end = max(fog_start + 5, far)
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glEnable(gl.GL_FOG)
            gl.glFogi(gl.GL_FOG_MODE, gl.GL_LINEAR)
            gl.glFogf(gl.GL_FOG_START, fog_start)
            gl.glFogf(gl.GL_FOG_END, fog_end)
            b = self.background_rgb
            gl.glFogfv(gl.GL_FOG_COLOR, [b[0], b[1], b[2], 1.0])
        else:
            gl.glDisable(gl.GL_FOG)

    def set_minimum_covering_sphere(self, atoms=None):
        if atoms is None:
            points = self.points
        else:
            points = flex.vec3_double()
            for atom in atoms:
                points.append(atom)
        if points is not None and len(points) > 1:
            s = scitbx.math.minimum_covering_sphere_3d(points=points)
        else:
            if points is None or len(points) == 0:
                center = (0, 0, 0)
            else:
                center = points[0]
            s = scitbx.math.sphere_3d(center=center, radius=1)
        self.minimum_covering_sphere = s

    def compute_home_translation(self):
        s = self.minimum_covering_sphere
        x, y, z = [-v for v in gltbx.util.object_as_eye_coordinates(s.center())]
        h = s.radius()
        if self.w < self.h:
            h *= self.h / max(1, self.w)
        assert 0 < self.field_of_view_y < 90
        z -= h / math.tan(self.field_of_view_y * math.pi / 180 / 2)
        return x, y, z

    def initialize_modelview(self, eye_vector=None, angle=None):
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.setup_lighting()
        glu.gluLookAt(0, 0, 0, 0, 0, -1, 0, 1, 0)
        translation = self.compute_home_translation()
        gl.glTranslated(*translation)
        rc = self.minimum_covering_sphere.center()
        self.rotation_center = rc
        if eye_vector is None:
            eye_vector = (1, 1, 1)
        if angle is None:
            angle = -120
        gltbx.util.rotate_object_about_eye_vector(
            xcenter=rc[0],
            ycenter=rc[1],
            zcenter=rc[2],
            xvector=eye_vector[0],
            yvector=eye_vector[1],
            zvector=eye_vector[2],
            angle=angle,
        )

    def rotation_move_factor(self, rotation_angle):
        return abs(rotation_angle) / 180

    def translation_move_factor(self, translation_vector):
        return min(
            2,
            abs(matrix.col(translation_vector))
            / max(1, self.minimum_covering_sphere.radius()),
        )

    def fit_into_viewport(self):
        dx, dy, dz = self.compute_home_translation()
        move_factor = self.translation_move_factor((dx, dy, dz))
        mvm = gltbx.util.get_gl_modelview_matrix()
        for f in animation_stepper(
            time_move=self.animation_time, move_factor=move_factor
        ):
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadIdentity()
            gl.glTranslated(f * dx, f * dy, f * dz)
            gl.glMultMatrixd(mvm)
            self.OnRedraw()

    def mark_rotation(self):
        self.marked_rotation = matrix.sqr(
            gltbx.util.extract_rotation_from_gl_modelview_matrix()
        )

    def snap_back_rotation(self):
        rc = self.rotation_center
        rotation_to_undo = matrix.sqr(
            gltbx.util.extract_rotation_from_gl_modelview_matrix()
        )
        if self.marked_rotation is not None:
            rotation_to_undo *= self.marked_rotation.inverse()
        aa = scitbx.math.r3_rotation_axis_and_angle_from_matrix(
            r=rotation_to_undo.as_mat3()
        )
        u, v, w = aa.axis
        angle = -aa.angle(deg=True)
        mvm = gltbx.util.get_gl_modelview_matrix()
        for f in animation_stepper(
            time_move=self.animation_time, move_factor=self.rotation_move_factor(angle)
        ):
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadMatrixd(mvm)
            gltbx.util.rotate_object_about_eye_vector(
                xcenter=rc[0],
                ycenter=rc[1],
                zcenter=rc[2],
                xvector=u,
                yvector=v,
                zvector=w,
                angle=f * angle,
            )
            self.OnRedraw()

    def move_rotation_center_to_mcs_center(self):
        self.rotation_center = self.minimum_covering_sphere.center()

    def move_to_center_of_viewport(self, obj_coor):
        dx, dy = [-x for x in gltbx.util.object_as_eye_coordinates(obj_coor)[:2]]
        move_factor = self.translation_move_factor((dx, dy, 0))
        mvm = gltbx.util.get_gl_modelview_matrix()
        for f in animation_stepper(
            time_move=self.animation_time, move_factor=move_factor
        ):
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadIdentity()
            gl.glTranslated(f * dx, f * dy, 0)
            gl.glMultMatrixd(mvm)
            self.OnRedraw()

    def ZoomOnSelection(self, xyzs):
        if not xyzs:
            return
        x, y, z = (0, 0, 0)
        for xyz in xyzs:
            x += xyz[0]
            y += xyz[1]
            z += xyz[2]
        x /= len(xyzs)
        y /= len(xyzs)
        z /= len(xyzs)
        self.rotation_center = (x, y, z)
        self.move_to_center_of_viewport(self.rotation_center)

    def OnRecordMouse(self, event):
        self.xmouse = event.GetX()
        self.ymouse = event.GetY()

    def OnStartRotate(self, event):
        self.autospin = False
        self.OnRecordMouse(event)

    def OnScale(self, scale):
        if abs(scale) > self.scale_max:
            if scale < 0:
                scale = -self.scale_max
            else:
                scale = self.scale_max
        s = self.minimum_covering_sphere
        r = (1 + 1.0e-6) * s.radius()
        d = -gltbx.util.object_as_eye_coordinates(self.rotation_center)[2]
        dr = d + r
        if scale > 0 and (dr <= self.min_near or d <= self.min_dist):
            pass  # near limit
        elif (
            scale < 0
            and r
            < d
            * math.sin(self.field_of_view_y * math.pi / 180 / 2)
            * self.min_viewport_use_fraction
        ):
            pass  # far limit
        else:
            gltbx.util.translate_object(0, 0, dr * scale)
            self.OnRedraw()

    def OnAutoSpin(self, event):
        if self.autospin_allowed:
            self.autospin = True
            self.yspin = 0.1 * (event.GetX() - self.initLeft[0])
            self.xspin = 0.1 * (event.GetY() - self.initLeft[1])
            if self.xspin == 0 and self.yspin == 0:
                self.autospin = False
            else:
                self.do_AutoSpin()

    def do_AutoSpin(self):
        spin_factor = 0.05
        rc = self.rotation_center
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gltbx.util.rotate_object_about_eye_x_and_y(
            spin_factor, rc[0], rc[1], rc[2], self.yspin, self.xspin, 0, 0
        )
        self.OnRedraw()

    def rotate_view(self, x1, y1, x2, y2, shift_down=False, scale=0.5):
        rc = self.rotation_center
        if not shift_down:
            gltbx.util.rotate_object_about_eye_x_and_y(
                scale, rc[0], rc[1], rc[2], x2, y2, x1, y1
            )
        else:
            sz = self.GetClientSize()
            sz = (sz[0] / 2, sz[1] / 2)
            dy = y1 - y2
            dx = x1 - x2
            if y2 > sz[1]:
                dx *= -1
            if x2 < sz[0]:
                dy *= -1
            angle = (dx + dy) / 2
            gltbx.util.rotate_object_about_eye_vector(
                xcenter=rc[0],
                ycenter=rc[1],
                zcenter=rc[2],
                xvector=0,
                yvector=0,
                zvector=1,
                angle=angle,
            )
        self.OnRedraw()

    def adjust_slab(self, delta):
        self.slab_scale += delta
        if self.slab_scale > 1.0:
            self.slab_scale = 1.0
        elif self.slab_scale < 0.01:
            self.slab_scale = 0.01

    def OnRotate(self, event):
        xp = event.GetX()
        yp = event.GetY()
        self.rotate_view(self.xmouse, self.ymouse, xp, yp, event.ShiftDown())
        self.OnRecordMouse(event)

    def OnTranslate(self, event):
        model = gltbx.util.get_gl_modelview_matrix()
        proj = gltbx.util.get_gl_projection_matrix()
        view = gltbx.util.get_gl_viewport()
        winx = []
        winy = []
        winz = []
        rc = self.rotation_center
        rc_eye = gltbx.util.object_as_eye_coordinates(rc)
        assert glu.gluProject(rc[0], rc[1], rc[2], model, proj, view, winx, winy, winz)
        objx = []
        objy = []
        objz = []
        win_height = max(1, self.w)
        assert glu.gluUnProject(
            winx[0],
            winy[0] + 0.5 * win_height,
            winz[0],
            model,
            proj,
            view,
            objx,
            objy,
            objz,
        )
        dist = v3distsq((objx[0], objy[0], objz[0]), rc) ** 0.5
        scale = abs(dist / (0.5 * win_height))
        x, y = event.GetX(), event.GetY()
        gltbx.util.translate_object(scale, x, y, self.xmouse, self.ymouse)
        self.rotation_center = tuple(
            gltbx.util.modelview_matrix_as_rt().inverse() * matrix.col(rc_eye)
        )
        self.OnRedraw()
        self.OnRecordMouse(event)

    def get_pick_points(self, mouse_xy):
        model = gltbx.util.get_gl_modelview_matrix()
        proj = gltbx.util.get_gl_projection_matrix()
        view = gltbx.util.get_gl_viewport()
        self.pick_points = []
        for winz in [0.0, 1.0]:
            objx = []
            objy = []
            objz = []
            ok = glu.gluUnProject(
                mouse_xy[0],
                self.h - mouse_xy[1],
                winz,
                model,
                proj,
                view,
                objx,
                objy,
                objz,
            )
            if not ok:
                self.pick_points = None
                break
            self.pick_points.append((objx[0], objy[0], objz[0]))

    def OnPaint(self, event=None):
        wx.PaintDC(self)
        if self.context is not None:
            self.SetCurrent(self.context)
        else:
            self.SetCurrent()
        if self.GL_uninitialised:
            gl.glViewport(0, 0, self.w, self.h)
            self.InitGL()
            self.GL_uninitialised = 0
        self.OnRedrawGL(event)

    def OnRedraw(self, event=None):
        wx.ClientDC(self)
        if self.context is not None:
            self.SetCurrent(self.context)
        else:
            self.SetCurrent()
        self.OnRedrawGL(event)

    def setup_distances(self):
        s = self.minimum_covering_sphere
        r = self.buffer_factor * s.radius()
        # z = -gltbx.util.object_as_eye_coordinates(s.center())[2]
        z = -gltbx.util.object_as_eye_coordinates(self.rotation_center)[2]
        if z < self.z_min:
            z = self.z_min
        self.near = max(self.min_near, z - r)
        self.far = max(self.near + 1, z + r)

    def OnRedrawGL(self, event=None):
        gltbx.util.handle_error()
        self.setup_distances()
        self.setup_viewing_volume()
        gltbx.util.handle_error()
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        gl.glClear(gl.GL_DEPTH_BUFFER_BIT)
        _ = gltbx.util.get_gl_modelview_matrix()
        self.DrawGL()
        gltbx.util.handle_error()
        gl.glFlush()
        self.SwapBuffers()
        if event is not None:
            event.Skip()

    def OnUpdate(self, event=None):
        pass

    def force_update(self, recenter=False):
        wx.PostEvent(self, ViewerUpdateEvent(data=None, recenter=recenter))

    def edit_opengl_settings(self, event=None):
        if self._settings_widget is None:
            self._settings_widget = OpenGLSettingsToolbox(self)
            self._settings_widget.Show()

    def save_screen_shot_via_pil(
        self, file_name="wx_viewer", extensions=["png", "jpg", "tiff", "eps", "pdf"]
    ):
        import gltbx.viewer_utils
        from libtbx.str_utils import show_string

        pil_image = gltbx.viewer_utils.read_pixels_to_pil_image(
            x=0, y=0, width=self.w, height=self.h
        )
        if pil_image is None:
            print(
                "Cannot save screen shot to file:"
                " Python Imaging Library (PIL) not available."
            )
            return 0
        print("Screen shot width x height: %d x %d" % (self.w, self.h))
        save = pil_image.save

        def try_save(file_name_ext):
            try:
                save(file_name_ext)
            except KeyboardInterrupt:
                raise
            except Exception:
                return False
            return True

        for ext in extensions:
            if file_name.endswith("." + ext):
                print(f"Writing file: {show_string(os.path.abspath(file_name))}")
                if not try_save(file_name_ext=file_name):
                    print(f"Failure saving screen shot as {ext.upper()} file.")
                return 1
        n_written = 0
        for ext in extensions:
            file_name_ext = file_name + "." + ext
            if not try_save(file_name_ext=file_name_ext):
                print(f"Image output format not available: {ext.upper()}")
            else:
                print(f"Wrote file: {show_string(os.path.abspath(file_name_ext))}")
                n_written += 1
        return n_written

    def save_screen_shot_via_gl2ps(self, file_name):
        from libtbx.str_utils import show_string

        gl2ps = gltbx.util.gl2ps_interface
        if not gl2ps(file_name=None, draw_background=False, callback=None):
            print(f"PDF output via gl2ps not available: cannot write file {file_name}")
            return 0
        try:
            # preempt potential error in C++, for better reporting here
            open(file_name, "wb")
        except KeyboardInterrupt:
            raise
        except Exception:
            print(
                "Error opening file for writing: %s"
                % show_string(os.path.abspath(file_name))
            )
            return 0
        gl2ps(file_name=file_name, draw_background=False, callback=self.OnRedraw)
        print(f"Wrote file: {show_string(os.path.abspath(file_name))}")
        return 1

    def save_screen_shot(
        self, file_name="wx_viewer", extensions=["png", "jpg", "tiff", "eps", "pdf"]
    ):
        extensions_pil = []
        save_pdf = file_name.endswith(".pdf")
        gl2ps_file_name = file_name
        if not save_pdf:
            for ext in extensions:
                if ext == "pdf":
                    save_pdf = True
                    gl2ps_file_name += "." + ext
                else:
                    extensions_pil.append(ext)
        n_written = 0
        if len(extensions_pil) != 0:
            n_written += self.save_screen_shot_via_pil(
                file_name=file_name, extensions=extensions_pil
            )
        if save_pdf:
            n_written += self.save_screen_shot_via_gl2ps(file_name=gl2ps_file_name)
        if n_written == 0:
            raise ValueError("Cannot save screen shot in any of the formats specified.")
        return n_written

    def OnClose(self, event=None):
        self.spintimer.Stop()


class show_points_and_lines_mixin(wxGLWindow):
    def __init__(self, *args, **keyword_args):
        wxGLWindow.__init__(self, *args, **keyword_args)
        self.buffer_factor = 2.0
        self.flag_show_minimum_covering_sphere = True
        self.flag_show_labels = True
        self.flag_show_points = True
        self.flag_show_lines = True
        self.flag_show_rotation_center = True
        self.flag_show_spheres = True
        self.long_labels = None
        self.labels = []
        self.points = flex.vec3_double()
        self.line_i_seqs = []
        self.line_colors = {}
        self.spheres = []
        self.line_width = 1
        self.labels_display_list = None
        self.points_display_list = None
        self.lines_display_list = None

    def InitGL(self):
        gltbx.util.handle_error()
        b = self.background_rgb
        gl.glClearColor(b[0], b[1], b[2], 0.0)
        self.minimum_covering_sphere_display_list = None
        self.initialize_modelview()
        gltbx.util.handle_error()

    def DrawGL(self):
        if self.flag_show_minimum_covering_sphere:
            self.draw_minimum_covering_sphere()
        if self.flag_show_points:
            self.draw_points()
        if self.flag_show_lines:
            self.draw_lines()
        if self.flag_show_rotation_center:
            self.draw_rotation_center()
        if self.flag_show_labels:
            self.draw_labels()
        if self.flag_show_spheres:
            self.draw_spheres()

    def draw_minimum_covering_sphere(self):
        if self.minimum_covering_sphere_display_list is None:
            self.minimum_covering_sphere_display_list = gltbx.gl_managed.display_list()
            self.minimum_covering_sphere_display_list.compile()
            s = self.minimum_covering_sphere
            c = s.center()
            r = s.radius()
            gray = 0.3
            gl.glColor3f(gray, gray, gray)
            gl.glBegin(gl.GL_POLYGON)
            for i in range(360):
                a = i * math.pi / 180
                rs = r * math.sin(a)
                rc = r * math.cos(a)
                gl.glVertex3f(c[0], c[1] + rs, c[2] + rc)
            gl.glEnd()
            self.draw_cross_at(c, color=(1, 0, 0))
            self.minimum_covering_sphere_display_list.end()
        self.minimum_covering_sphere_display_list.call()

    def draw_points(self):
        if self.points_display_list is None:
            self.points_display_list = gltbx.gl_managed.display_list()
            self.points_display_list.compile()
            gl.glLineWidth(1)
            for point in self.points:
                self.draw_cross_at(point)
            self.points_display_list.end()
        self.points_display_list.call()

    def draw_lines(self):
        if self.lines_display_list is None:
            self.lines_display_list = gltbx.gl_managed.display_list()
            self.lines_display_list.compile()
            assert self.line_width > 0
            for i_seqs in self.line_i_seqs:
                color = self.line_colors.get(tuple(i_seqs))
                if color is None:
                    color = self.line_colors.get(tuple(reversed(i_seqs)))
                    if color is None:
                        color = (1, 0, 1)
                gl.glColor3f(*color)
                gl.glLineWidth(self.line_width)
                gl.glBegin(gl.GL_LINES)
                gl.glVertex3f(*self.points[i_seqs[0]])
                gl.glVertex3f(*self.points[i_seqs[1]])
                gl.glEnd()
            self.lines_display_list.end()
        self.lines_display_list.call()

    def draw_labels(self, color=(1, 1, 1)):
        if self.labels_display_list is None:
            font = gltbx.fonts.ucs_bitmap_8x13
            font.setup_call_lists()
            self.labels_display_list = gltbx.gl_managed.display_list()
            self.labels_display_list.compile()
            gl.glColor3f(*color)
            for label, point in zip(self.labels, self.points):
                gl.glRasterPos3f(*point)
                font.render_string(label)
            self.labels_display_list.end()
        self.labels_display_list.call()

    def draw_cross_at(self, position_tuple, color=(1, 1, 1), f=0.1):
        (x, y, z) = position_tuple
        gl.glBegin(gl.GL_LINES)
        gl.glColor3f(*color)
        gl.glVertex3f(x - f, y, z)
        gl.glVertex3f(x + f, y, z)
        gl.glVertex3f(x, y - f, z)
        gl.glVertex3f(x, y + f, z)
        gl.glVertex3f(x, y, z - f)
        gl.glVertex3f(x, y, z + f)
        gl.glEnd()

    def draw_rotation_center(self):
        self.draw_cross_at(self.rotation_center, color=(0, 1, 0))

    def draw_spheres(self, solid=False):
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gray = 0.3
        gl.glColor3f(gray, gray, gray)
        if solid:
            gl.glEnable(gl.GL_LIGHTING)
            gl.glEnable(gl.GL_LIGHT0)
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_AMBIENT, [1, 1, 1, 1])
            gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, [0, 0, 1, 0])
            gl.glEnable(gl.GL_BLEND)
            gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
            gl.glMaterialfv(gl.GL_FRONT, gl.GL_DIFFUSE, [1, 1, 1, 0.5])
            sphere = gltbx.util.SolidSphere
            grid = 50
        else:
            sphere = gltbx.util.WireSphere
            grid = 20
        for i, (x, r) in enumerate(self.spheres):
            gl.glPushMatrix()
            gl.glTranslated(*(x))
            sphere(radius=r, slices=grid, stacks=grid)
            gl.glPopMatrix()
        if solid:
            gl.glDisable(gl.GL_LIGHTING)
            gl.glDisable(gl.GL_LIGHT0)
            gl.glDisable(gl.GL_BLEND)

    def process_pick_points(self):
        line = scitbx.math.line_given_points(self.pick_points)
        min_dist_sq = 1**2
        i_point_closest = None
        for i_point, point in enumerate(self.points):
            dist_sq = line.distance_sq(point=matrix.col(point))
            if min_dist_sq > dist_sq:
                min_dist_sq = dist_sq
                i_point_closest = i_point
        if i_point_closest is not None:
            self.rotation_center = self.points[i_point_closest]
            if self.long_labels is not None:
                assert len(self.long_labels) == len(self.points)
                lbl = self.long_labels[i_point_closest]
            elif self.labels is not None:
                assert len(self.labels) == len(self.points)
                lbl = self.labels[i_point_closest]
            else:
                lbl = "index %d" % i_point_closest
            txt = f"pick point: {lbl}"
            print(txt)
            self.parent.SetStatusText(txt)


class OpenGLSettingsToolbox(wx.MiniFrame):
    def __init__(self, parent):
        wx.MiniFrame.__init__(
            self,
            parent,
            -1,
            title="OpenGL settings",
            pos=(100, 100),
            style=wx.CAPTION | wx.CLOSE_BOX | wx.RAISED_BORDER,
        )
        self.parent = parent
        self.widgets = {}
        panel = wx.Panel(self, -1)
        main_sizer = wx.BoxSizer(wx.VERTICAL)
        fog_box = wx.CheckBox(panel, -1, "Use fog")
        fog_box.SetValue(parent.flag_show_fog)
        main_sizer.Add(fog_box, 0, wx.ALL, 5)
        self.fog_box = fog_box
        szr = wx.FlexGridSizer(rows=0, cols=2, vgap=5, hgap=5)
        main_sizer.Add(szr, 0, 0, 0)
        slab_label = wx.StaticText(panel, -1, "Slab:")
        slab_slider = wx.Slider(
            panel, -1, int(parent.slab_scale * 100), minValue=1, maxValue=100
        )
        szr.Add(slab_label, 0, wx.ALL, 5)
        szr.Add(slab_slider, 0, wx.ALL, 5)
        fog_label = wx.StaticText(panel, -1, "Fog scale:")
        fog_slider = wx.Slider(
            panel, -1, int(parent.fog_scale_factor * 100), minValue=1, maxValue=100
        )
        szr.Add(fog_label, 0, wx.ALL, 5)
        szr.Add(fog_slider, 0, wx.ALL, 5)
        self.widgets["slab_scale"] = slab_slider
        self.widgets["fog_scale_factor"] = fog_slider
        self.SetSizer(main_sizer)
        main_sizer.Fit(panel)
        self.Fit()
        self.Bind(wx.EVT_SLIDER, self.OnUpdate)
        self.Bind(wx.EVT_CHECKBOX, self.OnUpdate)
        self.Bind(wx.EVT_CLOSE, self.OnClose)

    def OnUpdate(self, event=None):
        for setting_name, widget in self.widgets.items():
            new_value = float(widget.GetValue()) / 100.0
            setattr(self.parent, setting_name, new_value)
        self.parent.flag_show_fog = self.fog_box.GetValue()
        self.parent.OnRedrawGL()

    def OnClose(self, event=None):
        self.Destroy()
        self.parent._settings_widget = None


class App(wx.App):
    def __init__(self, title="gltbx.wx_viewer", default_size=(600, 600)):
        self.title = title
        self.default_size = wx.Size(*default_size)
        wx.App.__init__(self, 0)

    def OnInit(self):
        self.frame = wx.Frame(
            None, -1, self.title, wx.DefaultPosition, self.default_size
        )
        self.frame.Bind(wx.EVT_CLOSE, self.OnFrameClose)

        self.frame.CreateStatusBar()

        tb = self.frame.CreateToolBar(
            style=wx.TB_HORIZONTAL | wx.NO_BORDER | wx.TB_FLAT | wx.TB_TEXT
        )
        isinstance(tb, wx.ToolBar)
        tb.SetToolBitmapSize((32, 32))

        import gltbx.wx_viewers_images as images

        tb.Bind(wx.EVT_TOOL, self.OnToolClick)

        self.mcs_center_id = wx.NewId()
        tb.AddSimpleTool(
            self.mcs_center_id,
            images.mcs_img.as_wx_Bitmap(),
            "Rotate around MCS center",
            "Object are to be rotated around the Minimum Covering Sphere (MCS)"
            "centre."
            " Keyboard shortcut: m",
        )

        self.center_on_screen_id = wx.NewId()
        tb.AddSimpleTool(
            self.center_on_screen_id,
            images.centre_img.as_wx_Bitmap(),
            "Center in window",
            "Shifts object so that centre of rotation is centered in window."
            " Keyboard shortcut: c",
        )

        self.fit_on_screen_id = wx.NewId()
        tb.AddSimpleTool(
            self.fit_on_screen_id,
            images.fit_img.as_wx_Bitmap(),
            "Fit in window",
            "Resizes and shifts object to fit into window." " Keyboard shortcut: f",
        )

        self.mark_snap_back_id = wx.NewId()
        tb.AddSimpleTool(
            self.mark_snap_back_id,
            images.mark_snap_back_img.as_wx_Bitmap(),
            "Mark orientation for snap-back",
            "Marks object orientation as the target of a subsequent snap-back."
            " Keyboard shortcut: k",
        )

        self.snap_back_id = wx.NewId()
        tb.AddSimpleTool(
            self.snap_back_id,
            images.snap_back_img.as_wx_Bitmap(),
            "Snap back orientation",
            "Rotates object back to the last marked orientation."
            " Keyboard shortcut: a",
        )

        self.toggle_spin_id = wx.NewId()
        tb.AddCheckTool(
            self.toggle_spin_id,
            images.spin_img.as_wx_Bitmap(),
            shortHelp="Spin on/off",
            longHelp="Turns auto-spin on/off. Keyboard shortcut: s",
        )

        tb.Realize()

        menuBar = wx.MenuBar()
        file_menu = wx.Menu()
        item = file_menu.Append(-1, "E&xit\tAlt-X", "Exit demo")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item)
        menuBar.Append(file_menu, "&File")

        self.frame.SetMenuBar(menuBar)
        self.init_view_objects()
        self.update_status_bar()
        self.view_objects.SetFocus()
        self.SetTopWindow(self.frame)
        self.frame.Show(True)
        return True

    def OnExitApp(self, event):
        self.frame.Close(True)

    def OnFrameClose(self, event):
        f = getattr(self.view_objects, "CleanupBeforeFrameClose", None)
        if f is not None:
            f()
        self.frame.Destroy()

    def OnToolClick(self, event):
        eid = event.GetId()
        if eid == self.mcs_center_id:
            self.view_objects.move_rotation_center_to_mcs_center()
        elif eid == self.center_on_screen_id:
            self.view_objects.move_to_center_of_viewport(
                self.view_objects.rotation_center
            )
        elif eid == self.fit_on_screen_id:
            self.view_objects.fit_into_viewport()
        elif eid == self.mark_snap_back_id:
            self.view_objects.mark_rotation()
        elif eid == self.snap_back_id:
            self.view_objects.snap_back_rotation()
        elif eid == self.toggle_spin_id:
            self.view_objects.autospin_allowed = not self.view_objects.autospin_allowed
            self.view_objects.autospin = False
            self.update_status_bar()
        else:
            raise RuntimeError("Unknown event Id: %d" % eid)

    def update_status_bar(self):
        self.frame.SetStatusText(
            f"Auto Spin {['Off', 'On'][int(self.view_objects.autospin_allowed)]}"
        )



"""
Test refinement of beam, detector and crystal orientation parameters
using generated reflection positions from ideal geometry.

Control of the experimental model and choice of minimiser is done via
PHIL, which means we can do, for example:

cctbx.python tst_orientation_refinement.py \
"random_seed=3; engine=LBFGScurvs"
"""


from __future__ import annotations

import sys


def test(args=[]):
    from math import pi

    from cctbx.sgtbx import space_group, space_group_symbols

    # Symmetry constrained parameterisation for the unit cell
    from cctbx.uctbx import unit_cell

    # We will set up a mock scan and a mock experiment list
    from dxtbx.model import ScanFactory
    from dxtbx.model.experiment_list import Experiment, ExperimentList
    from libtbx.phil import parse
    from libtbx.test_utils import approx_equal
    from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
    from scitbx import matrix
    from scitbx.array_family import flex

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )

    # Model parameterisations
    from dials.algorithms.refinement.parameterisation.detector_parameters import (
        DetectorParameterisationSinglePanel,
    )

    # Parameterisation of the prediction equation
    from dials.algorithms.refinement.parameterisation.prediction_parameters import (
        XYPhiPredictionParameterisation,
    )
    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )
    from dials.algorithms.refinement.reflection_manager import ReflectionManager

    # Imports for the target function
    from dials.algorithms.refinement.target import (
        LeastSquaresPositionalResidualWithRmsdCutoff,
    )

    # Reflection prediction
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    # Get modules to build models and minimiser using PHIL
    from . import geometry_phil, minimiser_phil, setup_geometry, setup_minimiser

    #############################
    # Setup experimental models #
    #############################

    master_phil = parse(f"{geometry_phil}\n{minimiser_phil}")

    models = setup_geometry.Extract(master_phil, cmdline_args=args)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 180 degree sequence
    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 1800),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(1800)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)
    im_width = myscan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, pi)
    assert approx_equal(im_width, 0.1 * pi / 180.0)

    # Build an experiment list
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    ###########################
    # Parameterise the models #
    ###########################

    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)

    # Fix beam to the X-Z plane (imgCIF geometry), fix wavelength
    s0_param.set_fixed([True, False, True])

    # Fix crystal parameters
    # xluc_param.set_fixed([True, True, True, True, True, True])

    ########################################################################
    # Link model parameterisations together into a parameterisation of the #
    # prediction equation                                                  #
    ########################################################################

    pred_param = XYPhiPredictionParameterisation(
        experiments, [det_param], [s0_param], [xlo_param], [xluc_param]
    )

    ################################
    # Apply known parameter shifts #
    ################################

    # shift detector by 1.0 mm each translation and 2 mrad each rotation
    det_p_vals = det_param.get_param_vals()
    p_vals = [a + b for a, b in zip(det_p_vals, [1.0, 1.0, 1.0, 2.0, 2.0, 2.0])]
    det_param.set_param_vals(p_vals)

    # shift beam by 2 mrad in free axis
    s0_p_vals = s0_param.get_param_vals()
    p_vals = list(s0_p_vals)

    p_vals[0] += 2.0
    s0_param.set_param_vals(p_vals)

    # rotate crystal a bit (=2 mrad each rotation)
    xlo_p_vals = xlo_param.get_param_vals()
    p_vals = [a + b for a, b in zip(xlo_p_vals, [2.0, 2.0, 2.0])]
    xlo_param.set_param_vals(p_vals)

    # change unit cell a bit (=0.1 Angstrom length upsets, 0.1 degree of
    # gamma angle)
    xluc_p_vals = xluc_param.get_param_vals()
    cell_params = mycrystal.get_unit_cell().parameters()
    cell_params = [a + b for a, b in zip(cell_params, [0.1, 0.1, 0.1, 0.0, 0.0, 0.1])]
    new_uc = unit_cell(cell_params)
    newB = matrix.sqr(new_uc.fractionalization_matrix()).transpose()
    S = symmetrize_reduce_enlarge(mycrystal.get_space_group())
    S.set_orientation(orientation=newB)
    X = tuple([e * 1.0e5 for e in S.forward_independent_parameters()])
    xluc_param.set_param_vals(X)

    #############################
    # Generate some reflections #
    #############################

    print("Reflections will be generated with the following geometry:")
    print(mybeam)
    print(mydetector)
    print(mycrystal)
    print("Target values of parameters are")
    msg = "Parameters: " + "%.5f " * len(pred_param)
    print(msg % tuple(pred_param.get_param_vals()))
    print()

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        mycrystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    print("Total number of reflections excited", len(obs_refs))

    # Take only those rays that intersect the detector
    intersects = ray_intersection(mydetector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * pi / 180.0
    px_size = mydetector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    print("Total number of observations made", len(obs_refs))

    ###############################
    # Undo known parameter shifts #
    ###############################

    s0_param.set_param_vals(s0_p_vals)
    det_param.set_param_vals(det_p_vals)
    xlo_param.set_param_vals(xlo_p_vals)
    xluc_param.set_param_vals(xluc_p_vals)

    print("Initial values of parameters are")
    msg = "Parameters: " + "%.5f " * len(pred_param)
    print(msg % tuple(pred_param.get_param_vals()))
    print()

    #####################################
    # Select reflections for refinement #
    #####################################

    refman = ReflectionManager(obs_refs, experiments)

    ##############################
    # Set up the target function #
    ##############################

    # The current 'achieved' criterion compares RMSD against 1/3 the pixel size and
    # 1/3 the image width in radians. For the simulated data, these are just made up

    mytarget = LeastSquaresPositionalResidualWithRmsdCutoff(
        experiments, ref_predictor, refman, pred_param, restraints_parameterisation=None
    )

    ################################
    # Set up the refinement engine #
    ################################

    refiner = setup_minimiser.Extract(
        master_phil, mytarget, pred_param, cmdline_args=args
    ).refiner

    print("Prior to refinement the experimental model is:")
    print(mybeam)
    print(mydetector)
    print(mycrystal)

    refiner.run()

    print()
    print("Refinement has completed with the following geometry:")
    print(mybeam)
    print(mydetector)
    print(mycrystal)


if __name__ == "__main__":
    test(sys.argv[1:])



"""Tests for dials.report.analysis module"""

from __future__ import annotations

from unittest import mock

import pytest

from cctbx import miller
from dxtbx.model import Crystal
from dxtbx.serialize import load

from dials.algorithms.scaling.scaling_library import (
    merging_stats_from_scaled_array,
    scaled_data_as_miller_array,
)
from dials.array_family import flex
from dials.report.analysis import (
    batch_dependent_properties,
    combined_table_to_batch_dependent_properties,
    i_sig_i_vs_batch,
    reflection_tables_to_batch_dependent_properties,
    rmerge_vs_batch,
    scales_vs_batch,
    table_1_summary,
)


@pytest.fixture
def example_crystal():
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 1.0],
        "space_group_hall_symbol": "-P 2yb",
    }
    crystal = Crystal.from_dict(exp_dict)
    return crystal


@pytest.fixture
def example_miller_set(example_crystal):
    """Generate an example miller set."""
    ms = miller.set(
        crystal_symmetry=example_crystal.get_crystal_symmetry(),
        indices=flex.miller_index([(1, 1, 1)] * 8 + [(2, 2, 2)]),
        anomalous_flag=False,
    )
    return ms


@pytest.fixture
def batch_array(example_miller_set):
    """Generate an example batch array."""
    batches_ = flex.double([1, 1, 1, 2, 2, 2, 3, 3, 4])
    batches = miller.array(example_miller_set, data=batches_)
    return batches


@pytest.fixture
def data_array(example_miller_set):
    """Generate an example data array."""
    data_ = flex.double([1.0 + (i * 0.1) for i in range(0, 9)])
    data = miller.array(example_miller_set, data=data_)
    return data


expected_results = {
    "bins": [1, 2, 3, 4],
    "svb": [1.1, 1.4, 1.65, 1.8],
    "isigivb": [1.1 / 2.0, 1.4 / 2.0, 1.65 / 2.0, 1.8 / 2.0],
    "rmergevb": [0.22727, 0.059524, 0.18182, 0.0],
}


def test_scales_vs_batch(batch_array, data_array):
    """Test the scales_vs_batch function."""
    bins, svb = scales_vs_batch(data_array, batch_array)
    assert bins == expected_results["bins"]
    assert svb == pytest.approx(expected_results["svb"], 1e-6)


def test_IsigI_vs_batch(batch_array, data_array):
    """Test the IsigI_vs_batch function."""
    with pytest.raises(AssertionError):
        bins, isigivb = i_sig_i_vs_batch(data_array, batch_array)

    Is = data_array.customized_copy()
    Is.set_sigmas(flex.double(9, 2.0))
    bins, isigivb = i_sig_i_vs_batch(Is, batch_array)
    assert bins == expected_results["bins"]
    assert isigivb == pytest.approx(expected_results["isigivb"], 1e-6)


def test_Rmerge_vs_batch(batch_array, data_array):
    """Test the Rmerge_vs_batch function."""
    bins, rmergevsb = rmerge_vs_batch(data_array, batch_array)
    assert bins == expected_results["bins"]
    assert rmergevsb == pytest.approx(expected_results["rmergevb"], 1e-4)


def test_reflections_to_batch_properties(
    data_array, example_miller_set, example_crystal
):
    """Test the helper functions that provide the batch properties from reflection
    tables and experiments."""
    # first make a reflection table.
    reflections = flex.reflection_table()
    reflections["intensity.scale.value"] = data_array.data() * flex.double(9, 2.0)
    reflections["inverse_scale_factor"] = flex.double(9, 2.0)
    reflections["intensity.scale.variance"] = flex.double(9, 4.0) * flex.double(9, 4.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0, 0, 0.1)] * 3 + [(0, 0, 1.1)] * 3 + [(0, 0, 2.1)] * 2 + [(0, 0, 3.1)]
    )
    reflections["miller_index"] = example_miller_set.indices()
    reflections["id"] = flex.int(9, 1)
    reflections.set_flags(flex.bool(9, True), reflections.flags.integrated)
    reflections.set_flags(flex.bool(9, True), reflections.flags.scaled)

    experiments = [mock.Mock()]
    experiments[0].scan.get_image_range.return_value = [1, 10]
    experiments[0].crystal = example_crystal
    experiments[0].beam.get_wavelength.return_value = 1

    (
        bins,
        rmerge,
        isigi,
        scalesvsbatch,
        batch_data,
    ) = reflection_tables_to_batch_dependent_properties(  # pylint: disable=unbalanced-tuple-unpacking
        [reflections], experiments
    )

    assert bins == expected_results["bins"]
    assert rmerge == pytest.approx(expected_results["rmergevb"], 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"], 1e-4)
    assert scalesvsbatch == pytest.approx([2.0] * 4, 1e-4)
    assert batch_data == [{"range": (1, 10), "id": 0}]

    # now try a two experiment dataset in a combined table.
    import copy

    reflections_2 = copy.deepcopy(reflections)
    reflections_2["id"] = flex.int(9, 2)
    reflections.extend(reflections_2)
    experiments = [mock.Mock(), mock.Mock()]
    experiments[0].scan.get_image_range.return_value = [1, 10]
    experiments[0].crystal = example_crystal
    experiments[0].beam.get_wavelength.return_value = 1
    experiments[1].scan.get_image_range.return_value = [1, 10]
    experiments[1].crystal = example_crystal
    experiments[1].beam.get_wavelength.return_value = 1

    (
        bins,
        rmerge,
        isigi,
        scalesvsbatch,
        batch_data,
    ) = combined_table_to_batch_dependent_properties(  # pylint: disable=unbalanced-tuple-unpacking
        reflections, experiments
    )

    assert bins == [1, 2, 3, 4, 101, 102, 103, 104]
    assert rmerge == pytest.approx(expected_results["rmergevb"] * 2, 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"] * 2, 1e-4)
    assert scalesvsbatch == pytest.approx([2.0] * 8, 1e-4)
    assert batch_data == [{"range": (1, 10), "id": 0}, {"range": (101, 110), "id": 1}]


def test_batch_dependent_properties(batch_array, data_array):
    """Test the interface function that manages the calculations."""
    Is = data_array.customized_copy()
    Is.set_sigmas(flex.double(9, 2.0))

    bins, rmerge, isigi, scalesvsbatch = batch_dependent_properties(
        batch_array, Is, scales=data_array
    )

    assert bins == expected_results["bins"]
    assert rmerge == pytest.approx(expected_results["rmergevb"], 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"], 1e-4)
    assert scalesvsbatch == pytest.approx(expected_results["svb"], 1e-4)

    # test for no scales given
    bins, rmerge, isigi, scalesvsbatch = batch_dependent_properties(batch_array, Is)

    assert bins == expected_results["bins"]
    assert rmerge == pytest.approx(expected_results["rmergevb"], 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"], 1e-4)
    assert scalesvsbatch is None

    # test for bad input - batches should reduce for all data
    Is.set_sigmas(flex.double([2.0] * 8 + [-1.0] * 1))
    bins, rmerge, isigi, scalesvsbatch = batch_dependent_properties(
        batch_array, Is, scales=data_array
    )

    assert bins == expected_results["bins"][0:3]
    assert rmerge == pytest.approx(expected_results["rmergevb"][0:3], 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"][0:3], 1e-4)
    assert scalesvsbatch == pytest.approx(expected_results["svb"][0:3], 1e-4)

    bins, rmerge, isigi, scalesvsbatch = batch_dependent_properties(batch_array, Is)

    assert bins == expected_results["bins"][0:3]
    assert rmerge == pytest.approx(expected_results["rmergevb"][0:3], 1e-4)
    assert isigi == pytest.approx(expected_results["isigivb"][0:3], 1e-4)
    assert scalesvsbatch is None

    # test for mismatched array sizes
    with pytest.raises(AssertionError):
        _ = batch_dependent_properties(batch_array, Is, data_array[0:-1])
    with pytest.raises(AssertionError):
        _ = batch_dependent_properties(batch_array[0:-1], Is)


def test_table_1_summary(dials_data):
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    expts = load.experiment_list(location / "scaled_20_25.expt", check_format=False)
    refls = flex.reflection_table.from_file(location / "scaled_20_25.refl")
    # Get a miller array of real data and calculate an iotbx.merging_statistics
    ma = scaled_data_as_miller_array([refls], expts)
    arr, anom = merging_stats_from_scaled_array(ma)

    # Test that something is returned in each case
    ### Case of overall statistics summary
    out = table_1_summary(arr, anom)
    assert out
    assert all(a in out for a in ("Overall", "Low", "High"))
    assert "Suggested" not in out
    ### Case of overall and suggested statistics summary (with anom)
    out = table_1_summary(arr, anom, arr, anom)
    assert out
    assert all(a in out for a in ("Overall", "Suggested", "Low", "High"))
    ### Case of no anomalous, but with suggested as well as overall.
    out = table_1_summary(arr, selected_statistics=arr)
    assert out
    assert all(a in out for a in ("Overall", "Suggested", "Low", "High"))



from __future__ import annotations

import logging
from collections import defaultdict
from math import floor, sqrt

from cctbx import crystal, miller

from dials.array_family import flex

logger = logging.getLogger("dials.command_line.compute_delta_cchalf")


class ResolutionBinner:
    """
    A class to bin the data by resolution
    """

    def __init__(self, unit_cell, dmin, dmax, nbins, output=True):
        """
        Initialise the binner

        :param unit_cell: A unit cell object
        :param dmin: The maximum resolution
        :param dmax: The minimum resolution
        :param nbins: The number of bins
        """
        if output:
            logger.info("Resolution bins")
        assert dmin < dmax
        dmin_inv_sq = 1.0 / dmin**2
        dmax_inv_sq = 1.0 / dmax**2
        self._unit_cell = unit_cell
        self._nbins = nbins
        self._xmin = dmax_inv_sq
        self._xmax = dmin_inv_sq
        assert self._xmin < self._xmax
        self._bin_range = self._xmax - self._xmin
        self._bin_size = self._bin_range / self._nbins
        self._bins = []
        for i in range(self._nbins):
            b0, b1 = (
                self._xmin + i * self._bin_size,
                self._xmin + (i + 1) * self._bin_size,
            )
            if output:
                logger.info("%d: %.3f, %.3f", i, sqrt(1 / b0), sqrt(1 / b1))
            self._bins.append((b0, b1))

    def nbins(self):
        """
        :returns: The number of bins
        """
        return self._nbins

    def index(self, h):
        """
        Get the bin index from the miller index

        :param h: The miller index
        :returns: The bin index
        """
        d = self._unit_cell.d(h)
        d2 = 1 / d**2
        bin_index = int(floor((d2 - self._xmin) / self._bin_size))
        if bin_index >= self._nbins:
            bin_index = self._nbins - 1
        if bin_index < 0:
            bin_index = 0
        return bin_index


class ReflectionSum:
    """
    A helper class to store sums of X and X**2
    """

    def __init__(self, sum_x=0, sum_x2=0, n=0):
        self.sum_x = sum_x
        self.sum_x2 = sum_x2
        self.n = n


class BinData:
    """
    A helper class to store mean and variance
    """

    def __init__(self):
        self.mean = []
        self.var = []


def compute_cchalf(mean, var):
    """
    Compute the CC 1/2 using the formular from Assmann, Brehm and Diederichs 2016

    :param mean: The list of mean intensities
    :param var: The list of variances on the half set of mean intensities
    :returns: The CC 1/2
    """
    assert len(mean) == len(var)
    n = len(mean)
    mean_of_means = sum(mean) / n
    sigma_e = sum(var) / n
    sigma_y = sum((m - mean_of_means) ** 2 for m in mean) / (n - 1)
    cchalf = (sigma_y - sigma_e) / (sigma_y + sigma_e)
    return cchalf


def compute_mean_cchalf_in_bins(bin_data):
    """
    Compute the mean cchalf averaged across resolution bins

    :param bin_data: The mean and variance in each bin
    :returns: The mean CC 1/2
    """
    mean_cchalf = 0
    count = 0
    for bin_i in bin_data:
        mean = bin_i.mean
        var = bin_i.var
        n = len(mean)
        if n > 1:
            cchalf = compute_cchalf(mean, var)
            mean_cchalf += n * cchalf
            count += n
    if count:
        mean_cchalf /= count
    return mean_cchalf


def compute_cchalf_from_reflection_sums(reflection_sums, binner):
    """
    Compute the CC 1/2 by computing the CC 1/2 in resolution bins and then
    computing the weighted mean of the binned CC 1/2 values
    """
    # Compute Mean and variance of reflection intensities
    bin_data = [BinData() for _ in range(binner.nbins())]
    for h in reflection_sums:
        sum_x = reflection_sums[h].sum_x
        sum_x2 = reflection_sums[h].sum_x2
        n = reflection_sums[h].n

        if n > 1:
            mean = sum_x / n
            var = (sum_x2 - (sum_x) ** 2 / n) / (n - 1)
            var = var / n
            index = binner.index(h)
            bin_data[index].mean.append(mean)
            bin_data[index].var.append(var)

    # Compute the mean cchalf in resolution bins
    return compute_mean_cchalf_in_bins(bin_data)


class PerGroupCChalfStatistics:
    def __init__(
        self,
        reflection_table,
        mean_unit_cell,
        space_group,
        d_min=None,
        d_max=None,
        n_bins=10,
    ):
        # here dataset is the sweep number, group is the group number for doing
        # the cc half analysis. May be the same as dataset if doing per dataset
        # stats, but can also be different if doing image groups.

        self.mean_unit_cell = mean_unit_cell
        self.space_group = space_group
        self.d_min = d_min
        self.d_max = d_max
        self._num_bins = n_bins
        self.reflection_table = reflection_table
        self._cchalf_mean = None
        self._cchalf = None

        required = ["miller_index", "intensity", "variance", "dataset", "group"]

        for r in required:
            if r not in reflection_table:
                raise KeyError(f"Column {r} not present in reflection table")
        # now do a prefiltering
        self.map_to_asu()
        self.d_filter()

        # Create the resolution bins
        if self.d_min is None:
            self.d_min = flex.min(self.reflection_table["d"])
        if self.d_max is None:
            self.d_max = flex.max(self.reflection_table["d"])
        self.binner = ResolutionBinner(mean_unit_cell, self.d_min, self.d_max, n_bins)

        self.reflection_sums = defaultdict(ReflectionSum)
        self.compute_overall_stats()

    def compute_overall_stats(self):
        # Create lookups for elements by miller index
        index_lookup = defaultdict(list)
        for i, h in enumerate(self.reflection_table["miller_index"]):
            index_lookup[h].append(i)

        # Compute the Overall Sum(X) and Sum(X^2) for each unique reflection

        for h in index_lookup:
            sel = flex.size_t(index_lookup[h])
            intensities = self.reflection_table["intensity"].select(sel)
            n = intensities.size()
            sum_x = flex.sum(intensities)
            sum_x2 = flex.sum(flex.pow2(intensities))
            self.reflection_sums[h] = ReflectionSum(sum_x, sum_x2, n)

        # Compute some numbers
        self._num_datasets = len(set(self.reflection_table["dataset"]))
        self._num_groups = len(set(self.reflection_table["group"]))
        self._num_reflections = self.reflection_table.size()
        self._num_unique = len(self.reflection_sums)

        logger.info(
            """
Summary of input data:
# Datasets: %s
# Groups: %s
# Reflections: %s
# Unique reflections: %s""",
            self._num_datasets,
            self._num_groups,
            self._num_reflections,
            self._num_unique,
        )

    def map_to_asu(self):
        """Map the miller indices to the ASU"""
        # d = flex.double([self.mean_unit_cell.d(h) for h in reflection_table["miller_index"]])
        cs = crystal.symmetry(self.mean_unit_cell, space_group=self.space_group)
        ms = miller.set(cs, self.reflection_table["miller_index"], anomalous_flag=False)
        ms_asu = ms.map_to_asu()
        self.reflection_table["miller_index"] = ms_asu.indices()
        self.reflection_table["d"] = ms_asu.d_spacings().data()

    def d_filter(self):
        """Filter on d_min, d_max"""
        if (not self.d_min) and (not self.d_max):
            return
        if self.d_min is not None:
            sel = self.reflection_table["d"] > self.d_min
            if self.d_max is not None:
                sel &= self.reflection_table["d"] < self.d_max
        else:
            sel = self.reflection_table["d"] < self.d_max
        self.reflection_table.select(sel)

    def run(self):
        """Compute the ΔCC½ for all the data"""
        self._cchalf_mean = compute_cchalf_from_reflection_sums(
            self.reflection_sums, self.binner
        )
        logger.info("CC 1/2 mean: %.3f", (100 * self._cchalf_mean))
        self._cchalf = self._compute_cchalf_excluding_each_group()

    def _compute_cchalf_excluding_each_group(self):
        #   self, reflection_sums, binner, miller_index, dataset, intensity
        # ):
        """
        Compute the CC 1/2 with an image excluded.

        For each image, update the sums by removing the contribution from the image
        and then compute the CC 1/2 of the remaining data
        """

        # Create a lookup table for each reflection by dataset i.e. group of images
        dataset_lookup = defaultdict(list)
        for i, b in enumerate(self.reflection_table["group"]):
            dataset_lookup[b].append(i)

        # Compute CC1/2 minus each dataset
        cchalf_i = {}
        for dataset in dataset_lookup:

            # Find all observations from this dataset and create a lookup based on
            # miller index
            index_lookup = defaultdict(list)
            for i in dataset_lookup[dataset]:
                index_lookup[self.reflection_table["miller_index"][i]].append(i)

            # Loop through all the reflections and remove the contribution of
            # reflections from the current dataset
            dataset_reflection_sums = defaultdict(ReflectionSum)
            for h in self.reflection_sums:
                sum_x = self.reflection_sums[h].sum_x
                sum_x2 = self.reflection_sums[h].sum_x2
                n = self.reflection_sums[h].n
                for i in index_lookup[h]:
                    sum_x -= self.reflection_table["intensity"][i]
                    sum_x2 -= self.reflection_table["intensity"][i] ** 2
                    n -= 1
                dataset_reflection_sums[h] = ReflectionSum(sum_x, sum_x2, n)

            # Compute the CC 1/2 without the reflections from the current dataset
            cchalf = compute_cchalf_from_reflection_sums(
                dataset_reflection_sums, self.binner
            )
            cchalf_i[dataset] = cchalf
            logger.info("CC 1/2 excluding group %d: %.3f", dataset, 100 * cchalf)

        return cchalf_i

    def num_datasets(self):
        """
        Return the number of datasets
        """
        return len(self._cchalf)

    def num_reflections(self):
        """
        Return the number of reflections
        """
        return self._num_reflections

    def num_unique(self):
        """
        Return the number of unique reflections
        """
        return self._num_unique

    def mean_cchalf(self):
        """
        Return the mean CC 1/2
        """
        return self._cchalf_mean

    def cchalf_i(self):
        """
        Return the CC 1/2 for each image excluded
        """
        return self._cchalf

    def delta_cchalf_i(self):
        """
        Return the ΔCC½ for each image excluded
        """
        return {k: self._cchalf_mean - v for k, v in self._cchalf.items()}



"""
Definitions of scaling models.

A scaling model is a collection of scaling model components with appropriate
methods to define how these are composed into one model.
"""

from __future__ import annotations

import logging

from libtbx import Auto, phil

from dials.algorithms.scaling.error_model.error_model import BasicErrorModel
from dials.algorithms.scaling.model.components.scale_components import (
    LinearDoseDecay,
    QuadraticDoseDecay,
    SHScaleComponent,
    SingleBScaleFactor,
    SingleScaleFactor,
)
from dials.algorithms.scaling.model.components.smooth_scale_components import (
    SmoothBScaleComponent1D,
    SmoothScaleComponent1D,
    SmoothScaleComponent2D,
    SmoothScaleComponent3D,
)
from dials.algorithms.scaling.plots import (
    plot_absorption_parameters,
    plot_absorption_plots,
    plot_array_absorption_plot,
    plot_array_decay_plot,
    plot_array_modulation_plot,
    plot_dose_decay,
    plot_relative_Bs,
    plot_smooth_scales,
)
from dials.algorithms.scaling.scaling_utilities import sph_harm_table
from dials.array_family import flex
from dials_scaling_ext import (
    calc_lookup_index,
    calc_theta_phi,
    create_sph_harm_lookup_table,
)

logger = logging.getLogger("dials")

import pkg_resources

base_model_phil_str = """\
correction.fix = None
    .type = strings
    .help = "If specified, this correction will not be refined in this scaling run"
"""

kb_model_phil_str = (
    """\
decay_correction = True
    .type = bool
    .help = "Option to turn off decay correction (for physical/array/KB
            default models)."
    .expert_level = 1
"""
    + base_model_phil_str
)

dose_decay_model_phil_str = (
    """\
scale_interval = 2.0
    .type = float(value_min=1.0)
    .help = "Rotation (phi) interval between model parameters for the scale"
            "component."
    .expert_level = 1
relative_B_correction = True
    .type = bool
    .help = "Option to turn off relative B correction."
    .expert_level = 1
decay_correction = True
    .type = bool
    .help = "Option to turn off decay correction."
    .expert_level = 1
share.decay = True
    .type = bool
    .help = "Share the decay model between sweeps."
    .expert_level = 1
resolution_dependence = *quadratic linear
    .type = choice
    .help = "Use a dose model that depends linearly or quadratically on 1/d"
    .expert_level = 1
absorption_correction = False
    .type = bool
    .help = "Option to turn on spherical harmonic absorption correction."
    .expert_level = 1
lmax = 4
    .type = int(value_min=2)
    .help = "Number of spherical harmonics to include for absorption"
            "correction, recommended to be no more than 6."
    .expert_level = 2
surface_weight = 1e6
    .type = float(value_min=0.0)
    .help = "Restraint weight applied to spherical harmonic terms in the"
            "absorption correction."
    .expert_level = 2
fix_initial = True
    .type = bool
    .help = "If performing full matrix minimisation, in the final cycle,"
            "constrain the initial parameter for more reliable parameter and"
            "scale factor error estimates."
    .expert_level = 2
"""
    + base_model_phil_str
)


physical_model_phil_str = (
    """\
scale_interval = auto
    .type = float(value_min=1.0)
    .help = "Rotation (phi) interval between model parameters for the scale"
            "component (auto scales interval depending on oscillation range)."
    .expert_level = 1
decay_correction = True
    .type = bool
    .help = "Option to turn off decay correction."
    .expert_level = 1
decay_interval = auto
    .type = float(value_min=1.0)
    .help = "Rotation (phi) interval between model parameters for the decay"
            "component (auto scales interval depending on oscillation range)."
    .expert_level = 1
decay_restraint = 1e-1
    .type = float(value_min=0.0)
    .help = "Weight to weakly restrain B-values to 0."
    .expert_level = 2
absorption_correction = auto
    .type = bool
    .help = "Option to turn off absorption correction (default True if oscillation > 60.0)."
    .expert_level = 1
absorption_level = low medium high
    .type = choice
    .help = "Expected degree of relative absorption for different scattering"
            "paths through the crystal(s). If an option is selected, the"
            "scaling model parameters lmax and surface_weight will be set to"
            "appropriate values."
            "Relative absorption increases as crystal size increases,"
            "increases as wavelength increases and is increased as the crystal"
            "dimensions become less equal (i.e. is higher for needle shaped"
            "crystals and zero for a spherical crystal)."
            "Definitions of the levels and approximate correction magnitude:"
            "low:    ~1%% relative absorption, expected for typical protein"
            "        crystals (containing no strongly absorbing atoms) on the"
            "        order of ~100um measured at ~1A wavelength."
            "medium: ~5%% relative absorption"
            "high:   >25%% relative absorption, e.g. for measurements at long"
            "        wavelength or crystals with high absorption from heavy atoms."
    .expert_level = 1
lmax = auto
    .type = int(value_min=2)
    .help = "Number of spherical harmonics to include for absorption"
            "correction, defaults to 4 if no absorption_level is chosen."
            "It is recommended that the value need be no more than 6."
    .expert_level = 1
surface_weight = auto
    .type = float(value_min=0.0)
    .help = "Restraint weight applied to spherical harmonic terms in the"
            "absorption correction. A lower restraint allows a higher amount"
            "of absorption correction. Defaults to 5e5 if no absorption_level"
            "is chosen."
    .expert_level = 1
share.absorption = False
    .type = bool
    .help = "If True, a common absorption correction is refined across all sweeps".
fix_initial = True
    .type = bool
    .help = "If performing full matrix minimisation, in the final cycle,"
            "constrain the initial parameter for more reliable parameter and"
            "scale factor error estimates."
    .expert_level = 2
"""
    + base_model_phil_str
)

array_model_phil_str = (
    """\
decay_correction = True
    .type = bool
    .help = "Option to turn off decay correction (a 2D grid of parameters as"
            "a function of rotation and resolution (d-value))."
    .expert_level = 1
decay_interval = 20.0
    .type = float(value_min=1.0)
    .help = "Rotation (phi) interval between model parameters for the decay"
            "and absorption corrections."
    .expert_level = 1
n_resolution_bins = 10
    .type = int(value_min=1)
    .help = "Number of resolution bins to use for the decay term."
    .expert_level = 1
absorption_correction = True
    .type = bool
    .help = "Option to turn off absorption correction (a 3D grid of"
            "parameters as a function of rotation angle, detector-x and"
            "detector-y position)."
    .expert_level = 1
n_absorption_bins = 3
    .type = int(value_min=1)
    .help = "Number of bins in each dimension (applied to both x and y) for"
            "binning the detector position for the absorption term of the"
            "array model."
    .expert_level = 1
modulation_correction = False
    .type = bool
    .help = "Option to turn on a detector correction for the array default"
            "model."
    .expert_level = 2
n_modulation_bins = 20
    .type = int(value_min=1)
    .help = "Number of bins in each dimension (applied to both x and y) for"
            "binning the detector position for the modulation correction."
    .expert_level = 2
"""
    + base_model_phil_str
)

autos = [Auto, "auto", "Auto"]


class ScalingModelBase:
    """Abstract base class for scaling models."""

    id_ = None

    def __init__(self, configdict, is_scaled=False):
        """Initialise the model with no components and a :obj:`configdict`."""
        if not configdict["corrections"]:
            raise ValueError("No model components created.")
        self._components = {}
        self._configdict = configdict
        self._is_scaled = is_scaled
        self._error_model = None
        self._fixed_components = []

    @property
    def is_scaled(self):
        """:obj:`bool`: Indicate whether this model has previously been refined."""
        return self._is_scaled

    def fix_initial_parameter(self, params):
        """Fix a parameter of the scaling model."""
        return False

    @property
    def fixed_components(self):
        return self._fixed_components

    @fixed_components.setter
    def fixed_components(self, components):
        self._fixed_components = components

    def limit_image_range(self, new_image_range):
        """Modify the model if necessary due to reducing the image range.

        Args:
            new_image_range (tuple): The (start, end) of the new image range.
        """
        pass

    def set_scaling_model_as_scaled(self):
        """Set the boolean 'is_scaled' flag as True."""
        self._is_scaled = True

    def set_scaling_model_as_unscaled(self):
        """Set the boolean 'is_scaled' flag as False."""
        self._is_scaled = False

    def configure_components(self, reflection_table, experiment, params):
        """Add the required reflection table data to the model components."""
        raise NotImplementedError()

    def plot_model_components(self, reflection_table=None):
        """Return a dict of plots for plotting model components with plotly."""
        return {}

    @classmethod
    def from_data(cls, params, experiment, reflection_table):
        """Create the model from input data."""
        raise NotImplementedError()

    def set_valid_image_range(self, image_range):
        """Set the valid image range for the model in the :obj:`configdict`."""
        self._configdict["valid_image_range"] = image_range

    @property
    def error_model(self):
        """:obj:`error_model`: The error model associated with the scaling model."""
        return self._error_model

    @property
    def configdict(self):
        """:obj:`dict`: a dictionary of the model configuration parameters."""
        return self._configdict

    @property
    def components(self):
        """:obj:`dict`: a dictionary of the model components."""
        return self._components

    @property
    def n_params(self):
        """:obj:`dict`: a dictionary of the model components."""
        return sum(c.parameters.size() for c in self._components.values())

    def consecutive_refinement_order(self):
        """:obj:`list`: a nested list of component names.

        This list indicates to the scaler the order to perform scaling in
        consecutive scaling mode.
        e.g. [['scale', 'decay'], ['absorption']] would cause the first cycle to
        refine scale and decay, and then absorption in a subsequent cycle.
        """
        raise NotImplementedError()

    def to_dict(self):
        """Serialize the model to a dictionary.

        Returns:
            dict: A dictionary representation of the model.
        """
        dictionary = {"__id__": self.id_}
        for key in self.components:
            dictionary[key] = {
                "n_parameters": self._components[key].n_params,
                "parameters": list(self._components[key].parameters),
                "null_parameter_value": self._components[key].null_parameter_value,
            }
            if self._components[key].parameter_esds:
                dictionary[key]["est_standard_devs"] = list(
                    self._components[key].parameter_esds
                )
        dictionary["configuration_parameters"] = self._configdict
        return dictionary

    @classmethod
    def from_dict(cls, obj):
        """Create a scaling model from a dictionary."""
        raise NotImplementedError()

    def update(self, model_params):
        pass

    def load_error_model(self, error_params):
        # load existing model if there, but use user-specified values if given
        new_model = None
        if (
            "error_model_type" in self._configdict
            and not error_params.reset_error_model
        ):
            if self._configdict["error_model_type"] == "BasicErrorModel":
                p = self._configdict["error_model_parameters"]
                a = None
                b = None
                if not error_params.basic.a:
                    a = p[0]
                if not error_params.basic.b:
                    b = p[1]
                new_model = BasicErrorModel(a, b, error_params.basic)
        if not new_model:
            new_model = BasicErrorModel(basic_params=error_params.basic)
        logger.info(f"Loaded error model: {new_model}")
        self.set_error_model(new_model)

    def set_error_model(self, error_model):
        """Associate an error model with the dataset."""
        self._error_model = error_model
        self._configdict.update(
            {
                "error_model_type": self.error_model.__class__.__name__,
                "error_model_parameters": list(error_model.parameters),
            }
        )

    def record_intensity_combination_Imid(self, Imid):
        """Record the intensity combination Imid value."""
        self._configdict["Imid"] = Imid

    def get_shared_components(self):
        return None

    def __str__(self):
        """:obj:`str`: Return a string representation of a scaling model."""
        msg = ["Scaling model:"]
        msg.append("  type : " + str(self.id_))
        for name, component in self.components.items():
            msg.append("  " + str(name).capitalize() + " component:")
            if component.parameter_esds:
                msg.append("    parameters (sigma)")
                for p, e in zip(component.parameters, component.parameter_esds):
                    if p < 0.0:
                        msg.append(f"    {p:.4f}   ({e:.4f})")
                    else:
                        msg.append(f"     {p:.4f}   ({e:.4f})")
            else:
                msg.append("    parameters")
                for p in component.parameters:
                    if p < 0.0:
                        msg.append(f"    {p:.4f}")
                    else:
                        msg.append(f"     {p:.4f}")
        msg.append("")
        return "\n".join(msg)


def _add_absorption_component_to_physically_derived_model(model, reflection_table):
    lmax = model.configdict["lmax"]
    if reflection_table.size() > 100000:
        assert "s0c" in reflection_table
        assert "s1c" in reflection_table
        theta_phi_0 = calc_theta_phi(
            reflection_table["s0c"]
        )  # array of tuples in radians
        theta_phi_1 = calc_theta_phi(reflection_table["s1c"])
        s0_lookup_index = calc_lookup_index(theta_phi_0, points_per_degree=2)
        s1_lookup_index = calc_lookup_index(theta_phi_1, points_per_degree=2)
        if SHScaleComponent.coefficients_list is None:
            SHScaleComponent.coefficients_list = create_sph_harm_lookup_table(
                lmax, points_per_degree=2
            )  # set the class variable and share
        elif len(SHScaleComponent.coefficients_list) < (lmax * (2.0 + lmax)):
            # this (rare) case can happen if adding a new dataset with a larger lmax!
            SHScaleComponent.coefficients_list = create_sph_harm_lookup_table(
                lmax, points_per_degree=2
            )  # set the class variable and share
        model.components["absorption"].data = {
            "s0_lookup": s0_lookup_index,
            "s1_lookup": s1_lookup_index,
        }
    # here just pass in good reflections
    else:
        model.components["absorption"].data = {
            "sph_harm_table": sph_harm_table(reflection_table, lmax)
        }
    surface_weight = model.configdict["abs_surface_weight"]
    parameter_restraints = flex.double([])
    for i in range(1, lmax + 1):
        parameter_restraints.extend(flex.double([1.0] * ((2 * i) + 1)))
    parameter_restraints *= surface_weight
    model.components["absorption"].parameter_restraints = parameter_restraints


class DoseDecay(ScalingModelBase):

    """A model similar to the physical model, where an exponential decay
    component is used plus a relative B-factor per sweep, with no absorption
    surface by default. Most suitable for multi-crystal datasets."""

    id_ = "dose_decay"

    phil_scope = phil.parse(dose_decay_model_phil_str)

    def __init__(self, parameters_dict, configdict, is_scaled=False):
        """Create the physical scaling model components."""
        super().__init__(configdict, is_scaled)
        if "scale" in configdict["corrections"]:
            scale_setup = parameters_dict["scale"]
            self._components["scale"] = SmoothScaleComponent1D(
                scale_setup["parameters"], scale_setup["parameter_esds"]
            )
        if "decay" in configdict["corrections"]:
            decay_setup = parameters_dict["decay"]
            if configdict["resolution_dependence"] == "linear":
                self._components["decay"] = LinearDoseDecay(
                    decay_setup["parameters"], decay_setup["parameter_esds"]
                )
            else:
                self._components["decay"] = QuadraticDoseDecay(
                    decay_setup["parameters"], decay_setup["parameter_esds"]
                )
        if "relative_B" in configdict["corrections"]:
            B_setup = parameters_dict["relative_B"]
            self._components["relative_B"] = SingleBScaleFactor(
                B_setup["parameters"], B_setup["parameter_esds"]
            )
        if "absorption" in configdict["corrections"]:
            absorption_setup = parameters_dict["absorption"]
            self._components["absorption"] = SHScaleComponent(
                absorption_setup["parameters"], absorption_setup["parameter_esds"]
            )

    @property
    def consecutive_refinement_order(self):
        """:obj:`list`: a nested list of component names to indicate scaling order."""
        return [["scale", "relative_B", "decay"], ["absorption"]]

    def fix_initial_parameter(self, params):
        if "scale" in self.components and params.dose_decay.fix_initial:
            self.components["scale"].fix_initial_parameter()
        return True

    def get_shared_components(self):
        if "shared" in self.configdict:
            if (
                "decay" in self.configdict["shared"]
                and "decay" not in self.fixed_components
            ):
                return "decay"
        return None

    def update(self, params):
        if params.dose_decay.correction.fix:
            self.fixed_components = params.dose_decay.correction.fix

    def configure_components(self, reflection_table, experiment, params):
        """Add the required reflection table data to the model components."""
        phi = reflection_table["xyzobs.px.value"].parts()[2]
        if "scale" in self.components:
            norm = phi * self._configdict["s_norm_fac"]
            self.components["scale"].data = {"x": norm}
        if "decay" in self.components:
            self.components["decay"].data = {"x": phi, "d": reflection_table["d"]}
        if "relative_B" in self.components:
            self.components["relative_B"].data = {"d": reflection_table["d"]}
        if "absorption" in self.components:
            _add_absorption_component_to_physically_derived_model(
                self, reflection_table
            )

    def limit_image_range(self, new_image_range):
        """Modify the model to be suitable for a reduced image range.

        For this model, this involves determining whether the number of parameters
        should be reduced and may reduce the number of parameters in the scale and
        decay components.

        Args:
            new_image_range (tuple): The (start, end) of the new image range.
        """
        conf = self.configdict
        current_image_range = conf["valid_image_range"]
        current_osc_range = conf["valid_osc_range"]
        # calculate one osc as don't have access to scan object here
        one_osc = (conf["valid_osc_range"][1] - conf["valid_osc_range"][0]) / (
            conf["valid_image_range"][1] - (conf["valid_image_range"][0] - 1)
        )
        new_osc_range = (
            (new_image_range[0] - current_image_range[0]) * one_osc,
            (new_image_range[1] - current_image_range[0] + 1) * one_osc,
        )
        if "scale" in self.components:
            n_param, s_norm_fac, scale_rot_int = initialise_smooth_input(
                new_osc_range, one_osc, conf["scale_rot_interval"]
            )
            n_old_params = len(self.components["scale"].parameters)
            conf["scale_rot_interval"] = scale_rot_int
            conf["s_norm_fac"] = s_norm_fac
            offset = calculate_new_offset(
                current_image_range[0],
                new_image_range[0],
                s_norm_fac,
                n_old_params,
                n_param,
            )
            new_params = self.components["scale"].parameters[offset : offset + n_param]
            self.components["scale"].set_new_parameters(new_params)

        new_osc_range_0 = current_osc_range[0] + (
            (new_image_range[0] - current_image_range[0]) * one_osc
        )
        new_osc_range_1 = current_osc_range[1] + (
            (new_image_range[1] - current_image_range[1]) * one_osc
        )
        self._configdict["valid_osc_range"] = (new_osc_range_0, new_osc_range_1)
        self.set_valid_image_range(new_image_range)

    @classmethod
    def from_data(cls, params, experiment, reflection_table):
        """Create the scaling model defined by the params."""

        params = params.dose_decay
        configdict = {"corrections": []}
        parameters_dict = {}

        osc_range = experiment.scan.get_oscillation_range()
        one_osc_width = experiment.scan.get_oscillation()[1]
        configdict.update({"valid_osc_range": osc_range})
        if params.share.decay:
            configdict.update({"shared": ["decay"]})

        configdict["corrections"].append("scale")
        n_scale_param, s_norm_fac, scale_rot_int = initialise_smooth_input(
            osc_range, one_osc_width, params.scale_interval
        )
        configdict.update(
            {"s_norm_fac": s_norm_fac, "scale_rot_interval": scale_rot_int}
        )
        parameters_dict["scale"] = {
            "parameters": flex.double(n_scale_param, 1.0),
            "parameter_esds": None,
        }

        if params.relative_B_correction:
            configdict["corrections"].append("relative_B")
            parameters_dict["relative_B"] = {
                "parameters": flex.double(1, 0.0),
                "parameter_esds": None,
            }

        if params.decay_correction:
            configdict["corrections"].append("decay")
            configdict.update({"resolution_dependence": params.resolution_dependence})
            parameters_dict["decay"] = {
                "parameters": flex.double(1, 0.0),
                "parameter_esds": None,
            }

        if params.absorption_correction:
            configdict["corrections"].append("absorption")
            n_abs_param = params.lmax * (
                2 + params.lmax
            )  # arithmetic sum formula (a1=3, d=2)
            configdict.update({"lmax": params.lmax})
            configdict.update({"abs_surface_weight": params.surface_weight})
            parameters_dict["absorption"] = {
                "parameters": flex.double(n_abs_param, 0.0),
                "parameter_esds": None,
            }

        model = cls(parameters_dict, configdict)
        if params.correction.fix:
            model.fixed_components = params.correction.fix
        return model

    @classmethod
    def from_dict(cls, obj):
        """Create a :obj:`PhysicalScalingModel` from a dictionary."""
        if obj["__id__"] != cls.id_:
            raise RuntimeError(f"expected __id__ {cls.id_}, got {obj['__id__']}")
        (s_params, d_params, abs_params, B) = (None, None, None, None)
        (s_params_sds, d_params_sds, a_params_sds, B_sd) = (None, None, None, None)
        configdict = obj["configuration_parameters"]
        if "scale" in configdict["corrections"]:
            s_params = flex.double(obj["scale"]["parameters"])
            if "est_standard_devs" in obj["scale"]:
                s_params_sds = flex.double(obj["scale"]["est_standard_devs"])
        if "relative_B" in configdict["corrections"]:
            B = flex.double(obj["relative_B"]["parameters"])
            if "est_standard_devs" in obj["relative_B"]:
                B_sd = flex.double(obj["relative_B"]["est_standard_devs"])
        if "decay" in configdict["corrections"]:
            d_params = flex.double(obj["decay"]["parameters"])
            if "est_standard_devs" in obj["decay"]:
                d_params_sds = flex.double(obj["decay"]["est_standard_devs"])
        if "absorption" in configdict["corrections"]:
            abs_params = flex.double(obj["absorption"]["parameters"])
            if "est_standard_devs" in obj["absorption"]:
                a_params_sds = flex.double(obj["absorption"]["est_standard_devs"])

        parameters_dict = {
            "scale": {"parameters": s_params, "parameter_esds": s_params_sds},
            "decay": {"parameters": d_params, "parameter_esds": d_params_sds},
            "relative_B": {"parameters": B, "parameter_esds": B_sd},
            "absorption": {"parameters": abs_params, "parameter_esds": a_params_sds},
        }

        return cls(parameters_dict, configdict, is_scaled=True)

    def plot_model_components(self, reflection_table=None):
        d = plot_dose_decay(self)
        if "absorption" in self.components:
            d.update(plot_absorption_parameters(self))
            d.update(plot_absorption_plots(self, reflection_table))
        return d


def determine_auto_absorption_params(absorption):
    if absorption == "high":
        lmax = 6
        surface_weight = 5e3
    elif absorption == "medium":
        lmax = 6
        surface_weight = 5e4
    else:  # low
        lmax = 4
        surface_weight = 5e5
    return lmax, surface_weight


class PhysicalScalingModel(ScalingModelBase):
    """A scaling model for a physical parameterisation."""

    id_ = "physical"

    phil_scope = phil.parse(physical_model_phil_str)

    def __init__(self, parameters_dict, configdict, is_scaled=False):
        """Create the physical scaling model components."""
        super().__init__(configdict, is_scaled)
        if "scale" in configdict["corrections"]:
            scale_setup = parameters_dict["scale"]
            self._components["scale"] = SmoothScaleComponent1D(
                scale_setup["parameters"], scale_setup["parameter_esds"]
            )
        if "decay" in configdict["corrections"]:
            decay_setup = parameters_dict["decay"]
            self._components["decay"] = SmoothBScaleComponent1D(
                decay_setup["parameters"], decay_setup["parameter_esds"]
            )
        if "absorption" in configdict["corrections"]:
            absorption_setup = parameters_dict["absorption"]
            self._components["absorption"] = SHScaleComponent(
                absorption_setup["parameters"], absorption_setup["parameter_esds"]
            )

    @property
    def consecutive_refinement_order(self):
        """:obj:`list`: a nested list of component names to indicate scaling order."""
        return [["scale", "decay"], ["absorption"]]

    def fix_initial_parameter(self, params):
        if "scale" in self.components and params.physical.fix_initial:
            self.components["scale"].fix_initial_parameter()
        return True

    def configure_components(self, reflection_table, experiment, params):
        """Add the required reflection table data to the model components."""
        phi = reflection_table["xyzobs.px.value"].parts()[2]
        if "scale" in self.components:
            norm = phi * self._configdict["s_norm_fac"]
            self.components["scale"].data = {"x": norm}
        if "decay" in self.components:
            norm = phi * self._configdict["d_norm_fac"]
            self.components["decay"].parameter_restraints = flex.double(
                self.components["decay"].parameters.size(),
                params.physical.decay_restraint,
            )
            self.components["decay"].data = {"x": norm, "d": reflection_table["d"]}
        if "absorption" in self.components:
            _add_absorption_component_to_physically_derived_model(
                self, reflection_table
            )

    def limit_image_range(self, new_image_range):
        """Modify the model to be suitable for a reduced image range.

        For this model, this involves determining whether the number of parameters
        should be reduced and may reduce the number of parameters in the scale and
        decay components.

        Args:
            new_image_range (tuple): The (start, end) of the new image range.
        """
        conf = self.configdict
        current_image_range = conf["valid_image_range"]
        current_osc_range = conf["valid_osc_range"]
        # calculate one osc as don't have access to scan object here
        one_osc = (conf["valid_osc_range"][1] - conf["valid_osc_range"][0]) / (
            conf["valid_image_range"][1] - (conf["valid_image_range"][0] - 1)
        )
        new_osc_range = (
            (new_image_range[0] - current_image_range[0]) * one_osc,
            (new_image_range[1] - current_image_range[0] + 1) * one_osc,
        )
        if "scale" in self.components:
            n_param, s_norm_fac, scale_rot_int = initialise_smooth_input(
                new_osc_range, one_osc, conf["scale_rot_interval"]
            )
            n_old_params = len(self.components["scale"].parameters)
            conf["scale_rot_interval"] = scale_rot_int
            conf["s_norm_fac"] = s_norm_fac
            offset = calculate_new_offset(
                current_image_range[0],
                new_image_range[0],
                s_norm_fac,
                n_old_params,
                n_param,
            )
            new_params = self.components["scale"].parameters[offset : offset + n_param]
            self.components["scale"].set_new_parameters(new_params)
        if "decay" in self.components:
            n_param, d_norm_fac, decay_rot_int = initialise_smooth_input(
                new_osc_range, one_osc, conf["decay_rot_interval"]
            )
            n_old_params = len(self.components["decay"].parameters)
            conf["decay_rot_interval"] = decay_rot_int
            conf["d_norm_fac"] = d_norm_fac
            offset = calculate_new_offset(
                current_image_range[0],
                new_image_range[0],
                d_norm_fac,
                n_old_params,
                n_param,
            )
            new_params = self.components["decay"].parameters[offset : offset + n_param]
            self.components["decay"].set_new_parameters(new_params)

        new_osc_range_0 = current_osc_range[0] + (
            (new_image_range[0] - current_image_range[0]) * one_osc
        )
        new_osc_range_1 = current_osc_range[1] + (
            (new_image_range[1] - current_image_range[1]) * one_osc
        )
        self._configdict["valid_osc_range"] = (new_osc_range_0, new_osc_range_1)
        self.set_valid_image_range(new_image_range)

    @classmethod
    def from_data(cls, params, experiment, reflection_table):
        """Create the scaling model defined by the params."""

        params = params.physical
        configdict = {"corrections": []}
        parameters_dict = {}

        osc_range = experiment.scan.get_oscillation_range()
        one_osc_width = experiment.scan.get_oscillation()[1]
        configdict.update({"valid_osc_range": osc_range})

        abs_osc_range = abs(osc_range[1] - osc_range[0])

        if params.scale_interval in autos or params.decay_interval in autos:
            if abs_osc_range < 5.0:
                scale_interval, decay_interval = (1.0, 1.5)
            elif abs_osc_range < 10.0:
                scale_interval, decay_interval = (2.0, 3.0)
            elif abs_osc_range < 25.0:
                scale_interval, decay_interval = (4.0, 5.0)
            elif abs_osc_range < 90.0:
                scale_interval, decay_interval = (8.0, 10.0)
            else:
                scale_interval, decay_interval = (15.0, 20.0)

        if params.scale_interval not in autos:
            scale_interval = params.scale_interval
        if params.decay_interval not in autos:
            decay_interval = params.decay_interval

        configdict["corrections"].append("scale")
        n_scale_param, s_norm_fac, scale_rot_int = initialise_smooth_input(
            osc_range, one_osc_width, scale_interval
        )
        configdict.update(
            {"s_norm_fac": s_norm_fac, "scale_rot_interval": scale_rot_int}
        )
        parameters_dict["scale"] = {
            "parameters": flex.double(n_scale_param, 1.0),
            "parameter_esds": None,
        }

        if params.decay_correction:
            configdict["corrections"].append("decay")
            n_decay_param, d_norm_fac, decay_rot_int = initialise_smooth_input(
                osc_range, one_osc_width, decay_interval
            )
            configdict.update(
                {"d_norm_fac": d_norm_fac, "decay_rot_interval": decay_rot_int}
            )
            parameters_dict["decay"] = {
                "parameters": flex.double(n_decay_param, 0.0),
                "parameter_esds": None,
            }
        if params.absorption_correction in autos:
            if abs_osc_range > 60.0:
                absorption_correction = True
            else:
                absorption_correction = False
        else:
            absorption_correction = params.absorption_correction
        if absorption_correction or params.absorption_level:
            configdict["corrections"].append("absorption")
            if params.share.absorption:
                configdict.update({"shared": ["absorption"]})
            if params.absorption_level:
                lmax, surface_weight = determine_auto_absorption_params(
                    params.absorption_level
                )
                if (params.lmax not in autos) or (params.surface_weight not in autos):
                    logger.info(
                        """Using lmax, surface_weight parameters set by the absorption_level option,
                        rather than user specified options"""
                    )
            else:
                lmax, surface_weight = (params.lmax, params.surface_weight)
                if lmax in autos:
                    lmax = 4
                if params.surface_weight in autos:
                    surface_weight = 5e5
            n_abs_param = (2 * lmax) + (lmax**2)  # arithmetic sum formula (a1=3, d=2)
            configdict.update({"lmax": lmax})
            configdict.update({"abs_surface_weight": surface_weight})
            parameters_dict["absorption"] = {
                "parameters": flex.double(n_abs_param, 0.0),
                "parameter_esds": None,
            }

        model = cls(parameters_dict, configdict)
        if params.correction.fix:
            model.fixed_components = params.correction.fix
        return model

    @classmethod
    def from_dict(cls, obj):
        """Create a :obj:`PhysicalScalingModel` from a dictionary."""
        if obj["__id__"] != cls.id_:
            raise RuntimeError(f"expected __id__ {cls.id_}, got {obj['__id__']}")
        (s_params, d_params, abs_params) = (None, None, None)
        (s_params_sds, d_params_sds, a_params_sds) = (None, None, None)
        configdict = obj["configuration_parameters"]
        if "scale" in configdict["corrections"]:
            s_params = flex.double(obj["scale"]["parameters"])
            if "est_standard_devs" in obj["scale"]:
                s_params_sds = flex.double(obj["scale"]["est_standard_devs"])
        if "decay" in configdict["corrections"]:
            d_params = flex.double(obj["decay"]["parameters"])
            if "est_standard_devs" in obj["decay"]:
                d_params_sds = flex.double(obj["decay"]["est_standard_devs"])
        if "absorption" in configdict["corrections"]:
            abs_params = flex.double(obj["absorption"]["parameters"])
            if "est_standard_devs" in obj["absorption"]:
                a_params_sds = flex.double(obj["absorption"]["est_standard_devs"])

        parameters_dict = {
            "scale": {"parameters": s_params, "parameter_esds": s_params_sds},
            "decay": {"parameters": d_params, "parameter_esds": d_params_sds},
            "absorption": {"parameters": abs_params, "parameter_esds": a_params_sds},
        }
        return cls(parameters_dict, configdict, is_scaled=True)

    def update(self, params):
        """Update the model if new options chosen in the phil scope."""
        if params.physical.correction.fix:
            self.fixed_components = params.physical.correction.fix
        if "absorption" in self.components:
            new_lmax = None
            if params.physical.absorption_level:
                lmax, surface_weight = determine_auto_absorption_params(
                    params.physical.absorption_level
                )
                self._configdict.update({"abs_surface_weight": surface_weight})
                if lmax != self._configdict["lmax"]:
                    new_lmax = lmax
            else:  # check manually specified parameters.
                if params.physical.surface_weight not in autos:
                    self._configdict.update(
                        {"abs_surface_weight": params.physical.surface_weight}
                    )
                if (params.physical.lmax not in autos) and (
                    params.physical.lmax != self._configdict["lmax"]
                ):
                    new_lmax = params.physical.lmax
            if new_lmax:
                # need to change the parameters for this component
                current_parameters = self.components["absorption"].parameters
                n_abs_param = new_lmax * (2 + new_lmax)
                self._configdict.update({"lmax": new_lmax})
                new_parameters = flex.double(n_abs_param, 0.0)
                # copy across matching parameters:
                for i in range(0, min(n_abs_param, current_parameters.size())):
                    new_parameters[i] = current_parameters[i]
                self._components["absorption"] = SHScaleComponent(
                    new_parameters, flex.double(n_abs_param, 0.0)
                )
            if params.physical.share.absorption:
                self._configdict.update({"shared": ["absorption"]})

    def plot_model_components(self, reflection_table=None):
        d = plot_smooth_scales(self)
        if "absorption" in self.components:
            d.update(plot_absorption_parameters(self))
            d.update(plot_absorption_plots(self, reflection_table))
        return d

    def get_shared_components(self):
        if "shared" in self.configdict:
            if (
                "absorption" in self.configdict["shared"]
                and "absorption" not in self.fixed_components
            ):
                return "absorption"
        return None


class ArrayScalingModel(ScalingModelBase):
    """A scaling model for an array-based parameterisation."""

    id_ = "array"

    phil_scope = phil.parse(array_model_phil_str)

    def __init__(self, parameters_dict, configdict, is_scaled=False):
        """Create the array scaling model components."""
        super().__init__(configdict, is_scaled)
        if not any(i in configdict["corrections"] for i in ["decay", "absorption"]):
            raise ValueError(
                "Array model must have at least one of decay or absorption corrections"
            )
        if "decay" in configdict["corrections"]:
            decay_setup = parameters_dict["decay"]
            self._components["decay"] = SmoothScaleComponent2D(
                decay_setup["parameters"],
                shape=(configdict["n_res_param"], configdict["n_time_param"]),
                parameter_esds=decay_setup["parameter_esds"],
            )
        if "absorption" in configdict["corrections"]:
            abs_setup = parameters_dict["absorption"]
            self._components["absorption"] = SmoothScaleComponent3D(
                abs_setup["parameters"],
                shape=(
                    configdict["n_x_param"],
                    configdict["n_y_param"],
                    configdict["n_time_param"],
                ),
                parameter_esds=abs_setup["parameter_esds"],
            )
        if "modulation" in configdict["corrections"]:
            mod_setup = parameters_dict["modulation"]
            self._components["modulation"] = SmoothScaleComponent2D(
                mod_setup["parameters"],
                shape=(configdict["n_x_mod_param"], configdict["n_y_mod_param"]),
                parameter_esds=mod_setup["parameter_esds"],
            )

    @property
    def consecutive_refinement_order(self):
        """:obj:`list`: a nested list of component names to indicate scaling order."""
        return [["decay"], ["absorption"], ["modulation"]]

    def update(self, params):
        if params.array.correction.fix:
            self.fixed_components = params.array.correction.fix

    def configure_components(self, reflection_table, experiment, params):
        """Add the required reflection table data to the model components."""
        xyz = reflection_table["xyzobs.px.value"].parts()
        norm_time = xyz[2] * self.configdict["time_norm_fac"]
        if "decay" in self.components:
            d = reflection_table["d"]
            norm_res = (
                (1.0 / flex.pow2(d)) - self.configdict["resmin"]
            ) / self.configdict["res_bin_width"]
            self.components["decay"].data = {"x": norm_res, "y": norm_time}
        if "absorption" in self.components:
            norm_x_abs = (xyz[0] - self.configdict["xmin"]) / self.configdict[
                "x_bin_width"
            ]
            norm_y_abs = (xyz[1] - self.configdict["ymin"]) / self.configdict[
                "y_bin_width"
            ]
            self.components["absorption"].data = {
                "x": norm_x_abs,
                "y": norm_y_abs,
                "z": norm_time,
            }
        if "modulation" in self.components:
            norm_x_det = (xyz[0] - self.configdict["xmin"]) / self.configdict[
                "x_det_bin_width"
            ]
            norm_y_det = (xyz[1] - self.configdict["ymin"]) / self.configdict[
                "y_det_bin_width"
            ]
            self.components["modulation"].data = {"x": norm_x_det, "y": norm_y_det}

    def limit_image_range(self, new_image_range):
        """Modify the model to be suitable for a reduced image range.

        For this model, this involves determining whether the number of parameters
        should be reduced and may reduce the number of parameters in the absorption
        and decay components.

        Args:
            new_image_range (tuple): The (start, end) of the new image range.
        """
        conf = self.configdict
        current_image_range = conf["valid_image_range"]
        current_osc_range = conf["valid_osc_range"]
        # calculate one osc as don't have access to scan object here
        one_osc = (conf["valid_osc_range"][1] - conf["valid_osc_range"][0]) / (
            conf["valid_image_range"][1] - (conf["valid_image_range"][0] - 1)
        )
        new_osc_range = (
            (new_image_range[0] - current_image_range[0]) * one_osc,
            (new_image_range[1] - current_image_range[0] + 1) * one_osc,
        )

        n_param, time_norm_fac, time_rot_int = initialise_smooth_input(
            new_osc_range, one_osc, conf["time_rot_interval"]
        )
        if "decay" in self.components:
            n_old_time_params = int(
                len(self.components["decay"].parameters)
                / self.components["decay"].n_x_params
            )
        else:
            n_old_time_params = int(
                len(self.components["absorption"].parameters)
                / (
                    self.components["absorption"].n_x_params
                    * self.components["absorption"].n_y_params
                )
            )
        offset = calculate_new_offset(
            current_image_range[0],
            new_image_range[0],
            time_norm_fac,
            n_old_time_params,
            n_param,
        )

        if "absorption" in self.components:
            params = self.components["absorption"].parameters
            n_x_params = self.components["absorption"].n_x_params
            n_y_params = self.components["absorption"].n_y_params
            # can't do simple slice as 3-dim array
            time_offset = offset * n_x_params * n_y_params
            new_params = params[
                time_offset : time_offset + (n_param * n_x_params * n_y_params)
            ]
            self.components["absorption"].set_new_parameters(
                new_params, shape=(n_x_params, n_y_params, n_param)
            )
        if "decay" in self.components:
            params = self.components["decay"].parameters
            n_decay_params = self.components["decay"].n_x_params
            # can't do simple slice as 2-dim array
            decay_offset = offset * n_decay_params
            new_params = params[
                decay_offset : decay_offset + (n_param * n_decay_params)
            ]
            self.components["decay"].set_new_parameters(
                new_params, shape=(n_decay_params, n_param)
            )
        self._configdict["n_time_param"] = n_param
        self._configdict["time_norm_fac"] = time_norm_fac
        self._configdict["time_rot_interval"] = time_rot_int
        new_osc_range_0 = current_osc_range[0] + (
            (new_image_range[0] - current_image_range[0]) * one_osc
        )
        new_osc_range_1 = current_osc_range[1] + (
            (new_image_range[1] - current_image_range[1]) * one_osc
        )
        self._configdict["valid_osc_range"] = (new_osc_range_0, new_osc_range_1)
        self.set_valid_image_range(new_image_range)

    @classmethod
    def from_data(cls, params, experiment, reflection_table):
        """create an array-based scaling model."""
        params = params.array
        reflections = reflection_table.select(reflection_table["d"] > 0.0)
        configdict = {"corrections": []}
        # First initialise things common to more than one correction.
        one_osc_width = experiment.scan.get_oscillation()[1]
        osc_range = experiment.scan.get_oscillation_range()
        configdict.update({"valid_osc_range": osc_range})
        n_time_param, time_norm_fac, time_rot_int = initialise_smooth_input(
            osc_range, one_osc_width, params.decay_interval
        )
        (xvalues, yvalues, _) = reflections["xyzobs.px.value"].parts()
        (xmax, xmin) = (flex.max(xvalues) + 0.001, flex.min(xvalues) - 0.001)
        (ymax, ymin) = (flex.max(yvalues) + 0.001, flex.min(yvalues) - 0.001)

        parameters_dict = {}

        if params.decay_correction:
            configdict["corrections"].append("decay")
            resmax = (1.0 / (flex.min(reflections["d"]) ** 2)) + 1e-6
            resmin = (1.0 / (flex.max(reflections["d"]) ** 2)) - 1e-6
            n_res_bins = params.n_resolution_bins
            n_res_param, res_bin_width = calc_n_param_from_bins(
                resmin, resmax, n_res_bins
            )
            configdict.update(
                {
                    "n_res_param": n_res_param,
                    "n_time_param": n_time_param,
                    "resmin": resmin,
                    "res_bin_width": res_bin_width,
                    "time_norm_fac": time_norm_fac,
                    "time_rot_interval": time_rot_int,
                }
            )
            parameters_dict["decay"] = {
                "parameters": flex.double((n_time_param * n_res_param), 1.0),
                "parameter_esds": None,
            }

        if params.absorption_correction:
            configdict["corrections"].append("absorption")
            nxbins = nybins = params.n_absorption_bins
            n_x_param, x_bin_width = calc_n_param_from_bins(xmin, xmax, nxbins)
            n_y_param, y_bin_width = calc_n_param_from_bins(ymin, ymax, nybins)
            configdict.update(
                {
                    "n_x_param": n_x_param,
                    "n_y_param": n_y_param,
                    "xmin": xmin,
                    "ymin": ymin,
                    "x_bin_width": x_bin_width,
                    "y_bin_width": y_bin_width,
                    "n_time_param": n_time_param,
                    "time_norm_fac": time_norm_fac,
                    "time_rot_interval": time_rot_int,
                }
            )
            parameters_dict["absorption"] = {
                "parameters": flex.double((n_x_param * n_y_param * n_time_param), 1.0),
                "parameter_esds": None,
            }

        if params.modulation_correction:
            configdict["corrections"].append("modulation")
            nx_det_bins = ny_det_bins = params.n_modulation_bins
            n_x_mod_param, x_det_bw = calc_n_param_from_bins(xmin, xmax, nx_det_bins)
            n_y_mod_param, y_det_bw = calc_n_param_from_bins(ymin, ymax, ny_det_bins)
            configdict.update(
                {
                    "n_x_mod_param": n_x_mod_param,
                    "n_y_mod_param": n_y_mod_param,
                    "xmin": xmin,
                    "ymin": ymin,
                    "x_det_bin_width": x_det_bw,
                    "y_det_bin_width": y_det_bw,
                }
            )
            parameters_dict["modulation"] = {
                "parameters": flex.double((n_x_mod_param * n_y_mod_param), 1.0),
                "parameter_esds": None,
            }

        model = cls(parameters_dict, configdict)
        if params.correction.fix:
            model.fixed_components = params.correction.fix
        return model

    @classmethod
    def from_dict(cls, obj):
        """Create an :obj:`ArrayScalingModel` from a dictionary."""
        if obj["__id__"] != cls.id_:
            raise RuntimeError(f"expected __id__ {cls.id_}, got {obj['__id__']}")
        configdict = obj["configuration_parameters"]
        (dec_params, abs_params, mod_params) = (None, None, None)
        (d_params_sds, a_params_sds, m_params_sds) = (None, None, None)
        if "decay" in configdict["corrections"]:
            dec_params = flex.double(obj["decay"]["parameters"])
            if "est_standard_devs" in obj["decay"]:
                d_params_sds = flex.double(obj["decay"]["est_standard_devs"])
        if "absorption" in configdict["corrections"]:
            abs_params = flex.double(obj["absorption"]["parameters"])
            if "est_standard_devs" in obj["absorption"]:
                a_params_sds = flex.double(obj["absorption"]["est_standard_devs"])
        if "modulation" in configdict["corrections"]:
            mod_params = flex.double(obj["modulation"]["parameters"])
            if "est_standard_devs" in obj["modulation"]:
                m_params_sds = flex.double(obj["modulation"]["est_standard_devs"])

        parameters_dict = {
            "decay": {"parameters": dec_params, "parameter_esds": d_params_sds},
            "absorption": {"parameters": abs_params, "parameter_esds": a_params_sds},
            "modulation": {"parameters": mod_params, "parameter_esds": m_params_sds},
        }

        return cls(parameters_dict, configdict, is_scaled=True)

    def plot_model_components(self, reflection_table=None):
        d = {}
        if "absorption" in self.components:
            d.update(plot_array_absorption_plot(self))
        if "decay" in self.components:
            d.update(plot_array_decay_plot(self))
        if "modulation" in self.components:
            d.update(plot_array_modulation_plot(self))
        return d


class KBScalingModel(ScalingModelBase):
    """A scaling model for a KB parameterisation."""

    id_ = "KB"

    phil_scope = phil.parse(kb_model_phil_str)

    def __init__(self, parameters_dict, configdict, is_scaled=False):
        """Create the KB scaling model components."""
        super().__init__(configdict, is_scaled)
        if "scale" in configdict["corrections"]:
            self._components["scale"] = SingleScaleFactor(
                parameters_dict["scale"]["parameters"],
                parameters_dict["scale"]["parameter_esds"],
            )
        if "decay" in configdict["corrections"]:
            self._components["decay"] = SingleBScaleFactor(
                parameters_dict["decay"]["parameters"],
                parameters_dict["decay"]["parameter_esds"],
            )

    def configure_components(self, reflection_table, experiment, params):
        """Add the required reflection table data to the model components."""
        if "scale" in self.components:
            self.components["scale"].data = {"id": reflection_table["id"]}
        if "decay" in self.components:
            self.components["decay"].data = {
                "d": reflection_table["d"],
            }

    @property
    def consecutive_refinement_order(self):
        """:obj:`list`: a nested list of component names to indicate scaling order."""
        return [["scale", "decay"]]

    def update(self, params):
        if params.KB.correction.fix:
            self.fixed_components = params.KB.correction.fix

    @classmethod
    def from_dict(cls, obj):
        """Create an :obj:`KBScalingModel` from a dictionary."""
        if obj["__id__"] != cls.id_:
            raise RuntimeError(f"expected __id__ {cls.id_}, got {obj['__id__']}")
        configdict = obj["configuration_parameters"]
        (s_params, d_params) = (None, None)
        (s_params_sds, d_params_sds) = (None, None)
        if "scale" in configdict["corrections"]:
            s_params = flex.double(obj["scale"]["parameters"])
            if "est_standard_devs" in obj["scale"]:
                s_params_sds = flex.double(obj["scale"]["est_standard_devs"])
        if "decay" in configdict["corrections"]:
            d_params = flex.double(obj["decay"]["parameters"])
            if "est_standard_devs" in obj["decay"]:
                d_params_sds = flex.double(obj["decay"]["est_standard_devs"])

        parameters_dict = {
            "scale": {"parameters": s_params, "parameter_esds": s_params_sds},
            "decay": {"parameters": d_params, "parameter_esds": d_params_sds},
        }

        return cls(parameters_dict, configdict, is_scaled=True)

    @classmethod
    def from_data(cls, params, experiment, reflection_table):
        """Create the :obj:`KBScalingModel` from data."""
        configdict = {"corrections": []}
        parameters_dict = {}

        if params.KB.decay_correction:
            configdict["corrections"].append("decay")
            parameters_dict["decay"] = {
                "parameters": flex.double([0.0]),
                "parameter_esds": None,
            }
        configdict["corrections"].append("scale")
        parameters_dict["scale"] = {
            "parameters": flex.double([1.0]),
            "parameter_esds": None,
        }

        model = cls(parameters_dict, configdict)
        if params.KB.correction.fix:
            model.fixed_components = params.KB.correction.fix
        return model


def calculate_new_offset(
    current_image_0, new_image_0, new_norm_fac, n_old_param, n_new_param
):
    """Calculate the parameter offset for the new image range.

    Returns:
        int: An offset to apply when selecting the new parameters from the
          existing parameters.
    """
    if n_old_param == 2:
        return 0  # can't have less than two params
    batch_difference = (new_image_0 - current_image_0) * new_norm_fac
    n_to_shift = int(batch_difference // 1)
    if batch_difference % 1 > 0.5:
        n_to_shift += 1
    return min(n_old_param - n_new_param, n_to_shift)  # can't shift by more
    # than difference between old and new


def initialise_smooth_input(osc_range, one_osc_width, interval):
    """Calculate the required smoother parameters.

    Using information about the sequence and the chosen parameterisation
    interval, the required parameters for the smoother are determined.

    Args:
        osc_range (tuple): The (start, stop) of an oscillation in degrees.
        one_osc_width (float): The oscillation width of a single image in degrees.
        interval (float): The required maximum separation between parameters
            in degrees.

    Returns:
        tuple: 3-element tuple containing;
            n_params (:obj:`int`): The number of parameters to use.
            norm_fac (:obj:`float`): The degrees to parameters space normalisation factor.
            interval (:obj:`float`): The actual interval in degrees between the parameters.
    """
    interval += 0.00001
    if (osc_range[1] - osc_range[0]) < (2.0 * interval):
        if (osc_range[1] - osc_range[0]) <= interval:
            rot_int = osc_range[1] - osc_range[0]
            n_param = 2
        else:
            rot_int = (osc_range[1] - osc_range[0]) / 2.0
            n_param = 3
    else:
        n_bins = max(int((osc_range[1] - osc_range[0]) / interval) + 1, 3)
        rot_int = (osc_range[1] - osc_range[0]) / float(n_bins)
        n_param = n_bins + 2
    norm_fac = 0.999 * one_osc_width / rot_int  # to make sure normalise values
    # fall within range of smoother.
    return n_param, norm_fac, rot_int


def calc_n_param_from_bins(value_min, value_max, n_bins):
    """Return the correct number of bins for initialising the gaussian
    smoothers."""
    assert n_bins > 0
    assert isinstance(n_bins, int)
    bin_width = (value_max - value_min) / n_bins
    if n_bins == 1:
        n_param = 2
    elif n_bins == 2:
        n_param = 3
    else:
        n_param = n_bins + 2
    return n_param, bin_width


model_phil_scope = phil.parse("")
_dxtbx_scaling_models = {
    ep.name: ep for ep in pkg_resources.iter_entry_points("dxtbx.scaling_model_ext")
}
assert (
    _dxtbx_scaling_models
), "No models registered with dxtbx.scaling_model_ext entry point"
model_phil_scope.adopt_scope(
    phil.parse(
        "model ="
        + " ".join(_dxtbx_scaling_models)
        + "\n    .type = choice"
        + "\n    .help = Set scaling model to be applied to input datasets"
        + "\n    .expert_level = 0"
    )
)
for entry_point_name, entry_point in _dxtbx_scaling_models.items():
    ext_master_scope = phil.parse("%s .expert_level=1 {}" % entry_point_name)
    ext_phil_scope = ext_master_scope.get_without_substitution(entry_point_name)
    assert len(ext_phil_scope) == 1
    ext_phil_scope = ext_phil_scope[0]
    ext_phil_scope.adopt_scope(entry_point.load().phil_scope)
    model_phil_scope.adopt_scope(ext_master_scope)


def plot_scaling_models(model, reflection_table=None):
    """Return a dict of component plots for the model for plotting with plotly."""
    return model.plot_model_components(reflection_table=reflection_table)


def make_combined_plots(data):
    """Make any plots that require evaluation of all models."""
    if all(d.id_ == "dose_decay" for d in data.values()):
        relative_Bs = [
            d.to_dict()["relative_B"]["parameters"][0] for d in data.values()
        ]
        return plot_relative_Bs(relative_Bs)
    return {}



from __future__ import annotations

import math

from scitbx import matrix
from scitbx.math import r3_rotation_axis_and_angle_from_matrix


def difference_rotation_matrix_axis_angle(crystal_a, crystal_b, target_angle=0):
    from cctbx import sgtbx

    # assert crystal_a.get_space_group() == crystal_b.get_space_group()
    space_group = crystal_b.get_space_group()
    best_R_ab = None
    best_cb_op = None
    best_axis = None
    best_angle = 1e8
    # iterate over space group ops to find smallest differences
    for i_op, op in enumerate(space_group.build_derived_laue_group().all_ops()):
        if op.r().determinant() < 0:
            continue
        elif not op.t().is_zero():
            continue
        cb_op = sgtbx.change_of_basis_op(op.inverse())
        crystal_b_sym = crystal_b.change_basis(cb_op)
        U_a = matrix.sqr(crystal_a.get_U())
        U_b = matrix.sqr(crystal_b_sym.get_U())
        assert U_a.is_r3_rotation_matrix()
        assert U_b.is_r3_rotation_matrix()
        # the rotation matrix to transform from U_a to U_b
        R_ab = U_b * U_a.transpose()
        axis_angle = r3_rotation_axis_and_angle_from_matrix(R_ab)
        axis = axis_angle.axis
        angle = axis_angle.angle() * 180.0 / math.pi
        for sign in (+1, -1):
            if abs(sign * angle - target_angle) < abs(best_angle - target_angle):
                best_angle = sign * angle
                best_axis = tuple(sign * a for a in axis)
                best_R_ab = R_ab if sign > 0 else R_ab.inverse()
                best_cb_op = cb_op if sign > 0 else cb_op.inverse()

    return best_R_ab, best_axis, best_angle, best_cb_op


def rotation_matrix_differences(
    crystal_models, miller_indices=None, comparison="pairwise"
):
    assert comparison in ("pairwise", "sequential"), comparison
    output = []
    for i, cm_i in enumerate(crystal_models):
        for j in range(i + 1, len(crystal_models)):
            if comparison == "sequential" and j > i + 1:
                break
            R_ij, axis, angle, cb_op = difference_rotation_matrix_axis_angle(
                cm_i, crystal_models[j]
            )
            output.append(f"Change of basis op: {cb_op}")
            output.append(
                "Rotation matrix to transform crystal %i to crystal %i:"
                % (i + 1, j + 1)
            )
            output.append(R_ij.mathematica_form(format="%.3f", one_row_per_line=True))
            output.append(
                f"Rotation of {angle:.3f} degrees about axis ({axis[0]:.3f}, {axis[1]:.3f}, {axis[2]:.3f})"
            )
            if miller_indices is not None:
                for hkl in miller_indices:
                    cm_j = crystal_models[j].change_basis(cb_op)
                    A_i = cm_i.get_A()
                    A_j = cm_j.get_A()
                    a_star_i = matrix.col(A_i[0:3])
                    b_star_i = matrix.col(A_i[3:6])
                    c_star_i = matrix.col(A_i[6:9])
                    a_star_j = matrix.col(A_j[0:3])
                    b_star_j = matrix.col(A_j[3:6])
                    c_star_j = matrix.col(A_j[6:9])

                    v_i = hkl[0] * a_star_i + hkl[1] * b_star_i + hkl[2] * c_star_i
                    v_j = hkl[0] * a_star_j + hkl[1] * b_star_j + hkl[2] * c_star_j

                    output.append(
                        f"({hkl[0]},{hkl[1]},{hkl[2]}): %.2f deg"
                        % v_i.angle(v_j, deg=True)
                    )
            output.append("")
    return "\n".join(output)



from __future__ import annotations


class html_report:
    def __init__(self, external_dependencies="remote"):
        self._content = []
        self.external_dependencies = external_dependencies

    def header(self):

        assert self.external_dependencies in ("remote", "local", "embed")

        if self.external_dependencies == "remote":

            plotly_js = (
                '<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>'
            )
            bootstrap_js = '<script src="http://maxcdn.bootstrapcdn.com/bootstrap/3.3.5/js/bootstrap.min.js"></script>'
            jquery_js = (
                '<script src="https://code.jquery.com/jquery-1.12.0.min.js"></script>'
            )
            bootstrap_css = '<link rel="stylesheet" href="http://maxcdn.bootstrapcdn.com/bootstrap/3.3.5/css/bootstrap.min.css">'
            katex_js = '<script type="text/javascript" src=https://cdnjs.cloudflare.com/ajax/libs/KaTeX/0.5.1/katex.min.js></script>'
            katex_auto_render_js = '<script type="text/javascript" src=https://cdnjs.cloudflare.com/ajax/libs/KaTeX/0.5.1/contrib/auto-render.min.js></script>'
            katex_css = '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/KaTeX/0.5.1/katex.min.css">'

        elif self.external_dependencies == "local":
            import libtbx.load_env

            css_dir = libtbx.env.find_in_repositories("dials/static/css")
            js_dir = libtbx.env.find_in_repositories("dials/static/js")
            katex_dir = libtbx.env.find_in_repositories("dials/static/katex")

            plotly_js = f'<script src="{js_dir}/plotly-latest.min.js"></script>'
            bootstrap_js = f'<script src="{js_dir}/bootstrap.min.js"></script>'
            jquery_js = f'<script src="{js_dir}/jquery-1.12.0.min.js"></script>'
            bootstrap_css = (
                f'<link rel="stylesheet" href="{css_dir}/bootstrap.min.css">'
            )
            katex_js = (
                f'<script type="text/javascript" src={katex_dir}/katex.min.js></script>'
            )
            katex_auto_render_js = (
                '<script type="text/javascript" src=%s/contrib/auto-render.min.js></script>'
                % katex_dir
            )
            katex_css = f'<link rel="stylesheet" href="{katex_dir}/katex.min.css">'

        elif self.external_dependencies == "embed":
            import os

            import libtbx.load_env

            css_dir = libtbx.env.find_in_repositories("dials/static/css")
            js_dir = libtbx.env.find_in_repositories("dials/static/js")
            katex_dir = libtbx.env.find_in_repositories("dials/static/katex")

            plotly_js = (
                "<script>%s</script>"
                % open(os.path.join(js_dir, "plotly-latest.min.js")).read()
            )
            bootstrap_js = (
                "<script>%s</script>"
                % open(os.path.join(js_dir, "bootstrap.min.js")).read()
            )
            jquery_js = (
                "<script>%s</script>"
                % open(os.path.join(js_dir, "jquery-1.12.0.min.js")).read()
            )
            bootstrap_css = (
                '<style type="text/css">%s</style>'
                % open(os.path.join(css_dir, "bootstrap.min.css")).read()
            )
            katex_js = (
                "<script>%s</script>"
                % open(os.path.join(katex_dir, "katex.min.js")).read()
            )
            katex_auto_render_js = (
                "<script>%s</script>"
                % open(os.path.join(katex_dir, "contrib/auto-render.min.js")).read()
            )
            katex_css = (
                '<style type="text/css">%s</style>'
                % open(os.path.join(katex_dir, "katex.min.css")).read()
            )

        html_header = """
<!DOCTYPE html>
<head>

<!-- Definitions for compatibility with the ccp4i2 browser -->
<script>
    if (!Function.prototype.bind) {{
        Function.prototype.bind = function(oThis) {{
            if (typeof this !== 'function') {{
                // closest thing possible to the ECMAScript 5
                // internal IsCallable function
                throw new TypeError('Function.prototype.bind - what is trying to be bound is not callable');
            }}

            var aArgs   = Array.prototype.slice.call(arguments, 1),
            fToBind = this,
            fNOP    = function() {{}},
            fBound  = function() {{
                return fToBind.apply(this instanceof fNOP
                                     ? this
                                     : oThis,
                                     aArgs.concat(Array.prototype.slice.call(arguments)));
            }};

            if (this.prototype) {{
                // Function.prototype doesn't have a prototype property
                fNOP.prototype = this.prototype;
            }}
            fBound.prototype = new fNOP();

            return fBound;
        }};
    }}
    if (typeof Float64Array === 'undefined') Float64Array = Float32Array;
</script>

<!-- Plotly.js -->
{plotly_js}

<meta name="viewport" content="width=device-width, initial-scale=1" charset="UTF-8">

{jquery_js}
{bootstrap_js}
{katex_js}
{katex_auto_render_js}
{katex_css}
{bootstrap_css}
<style type="text/css">
{css}
</style>

</head>
""".format(
            plotly_js=plotly_js,
            bootstrap_js=bootstrap_js,
            bootstrap_css=bootstrap_css,
            jquery_js=jquery_js,
            katex_js=katex_js,
            katex_auto_render_js=katex_auto_render_js,
            katex_css=katex_css,
            css=self.css(),
        )

        return html_header

    def css(self):

        return """
body {
  /*font-family: Helmet, Freesans, Helvetica, Arial, sans-serif;*/
  margin: 8px;
  min-width: 240px;
  margin-left: 5%;
  margin-right: 5%;
}

.plot {
  float: left;
  width: 600px;
  height: 400px;
  margin-bottom: 20px;
}

.katex-display {
  text-align: left;
}

}"""

    def body(self):
        html_body = """

<body>
  %s
  <script>
    renderMathInElement(
        document.body);
  </script>
</body>
""" % (
            "\n".join(content.html() for content in self._content)
        )

        return html_body

    def html(self):
        return "\n".join((self.header(), self.body()))

    def add_content(self, content):
        self._content.append(content)


class page_header:
    def __init__(self, title):
        self._title = title

    def html(self):
        html = f"""<div class="page-header">
  <h1>{self._title}</h1>
</div>"""
        return html


class panel_group:
    def __init__(self, panels):
        self._panels = panels

    def html(self):
        html = """
<div class="panel-group">
  %s
</div>
""" % (
            "\n".join(p.html() for p in self._panels)
        )
        return html


class panel:
    def __init__(self, title, panel_id, show=False):
        self._title = title
        self._panel_id = panel_id
        self._content = []
        self._show = show

    def add_content(self, content):
        self._content.append(content)

    def html(self):
        html = """
<div class="panel panel-default">
  <div class="panel-heading" data-toggle="collapse" href="#collapse_%(panel_id)s">
    <h4 class="panel-title">
      <a>%(title)s</a>
    </h4>
  </div>
  <div id="collapse_%(panel_id)s" class="panel-collapse collapse%(in)s">
    <div class="panel-body">
      %(content)s
    </div>
  </div>
</div>
""" % {
            "panel_id": self._panel_id,
            "title": self._title,
            "in": " in" if self._show else "",
            "content": "\n".join(content.html() for content in self._content),
        }
        return html


class table_responsive:
    def __init__(self, table_html, width=None):
        self._table_html = table_html
        self._width = width

    def html(self):
        if self._width is not None:
            style = ' style="width: %ipx"' % self._width
        else:
            style = ""
        html = f"""
<div class="table-responsive"{style}>
  {self._table_html}
</div>
"""
        return html


class plotly_graph:
    def __init__(self, json_data, div_id):
        self._json_data = json_data
        self._div_id = div_id

    def javascript(self):
        import json

        json_str = json.dumps(self._json_data)
        javascript = """
  <script>
    var graphs_{div_id} = {json};
    Plotly.newPlot({div_id}, graphs_{div_id}.data, graphs_{div_id}.layout);
  </script>
  """.format(
            div_id=self._div_id,
            json=json_str,
        )
        return javascript

    def html(self):

        return "\n".join(
            (
                '<div class="col-xs-6 col-sm-6 col-md-4 plot" id="%s"></div>'
                % self._div_id,
                self.javascript(),
            )
        )


class container_fluid:
    def __init__(self):
        self._content = []

    def add_content(self, content):
        self._content.append(content)

    def html(self):
        html = """
<div class="container-fluid">
%s
</div>
""" % "\n".join(
            content.html() for content in self._content
        )
        return html


class div:
    def __init__(self):
        self._content = []

    def add_content(self, content):
        self._content.append(content)

    def html(self):
        html = """
<div>
%s
</div>
""" % "\n".join(
            content.html() for content in self._content
        )
        return html


class raw_html:
    def __init__(self, raw_html):
        self._raw_html = raw_html

    def html(self):
        return self._raw_html



from __future__ import annotations

from dials_algorithms_profile_model_gaussian_rs_transform_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "CoordinateGenerator",
    "GridIndexGenerator",
    "MapFramesForward",
    "MapFramesReverse",
    "TransformForward",
    "TransformForwardNoModel",
    "TransformReverse",
    "TransformReverseNoModel",
    "TransformSpec",
    "beam_vector_map",
)



from __future__ import annotations

import math
import random

from dxtbx.serialize import load
from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.profile_model.gaussian_rs import (
    BBoxCalculator3D,
    CoordinateSystem,
)
from dials.algorithms.profile_model.gaussian_rs.transform import (
    MapFramesForward,
    MapFramesReverse,
)


def test_map_frames_forward(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )

    # Get the models
    beam = sequence.get_beam()
    detector = sequence.get_detector()
    gonio = sequence.get_goniometer()
    scan = sequence.get_scan()

    # Set the delta_divergence/mosaicity
    n_sigma = 3
    sigma_divergence = 0.060 * math.pi / 180
    mosaicity = 0.154 * math.pi / 180
    delta_divergence = n_sigma * sigma_divergence
    delta_mosaicity = n_sigma * mosaicity

    # Set the grid size
    grid_size = (4, 4, 4)

    # Create the E3 fraction object
    transform = MapFramesForward(
        scan.get_array_range()[0],
        scan.get_oscillation(deg=False)[0],
        scan.get_oscillation(deg=False)[1],
        mosaicity,
        n_sigma,
        grid_size[2],
    )

    # Create the bounding box calculator
    calculate_bbox = BBoxCalculator3D(
        beam, detector, gonio, scan, delta_divergence, delta_mosaicity
    )

    assert len(detector) == 1
    s0 = beam.get_s0()
    m2 = gonio.get_rotation_axis()
    s0_length = matrix.col(beam.get_s0()).length()

    for i in range(100):

        # Get random x, y, z
        x = random.uniform(0, 2000)
        y = random.uniform(0, 2000)
        z = random.uniform(0, 9)

        # Get random s1, phi, panel
        s1 = matrix.col(detector[0].get_pixel_lab_coord((x, y))).normalize() * s0_length
        phi = scan.get_angle_from_array_index(z, deg=False)
        panel = 0

        # Calculate the bounding box
        bbox = calculate_bbox(s1, z, panel)

        # Create the XDS coordinate system
        xcs = CoordinateSystem(m2, s0, s1, phi)

        # Calculate the transform fraction
        fraction = transform(bbox[4:], phi, xcs.zeta())

        # Ensure the minimum and maximum are 0 < 1
        fmax = flex.max(fraction)
        fmin = flex.min(fraction)
        assert fmax <= (1.0 + 5e-15) and fmax > 0.0, f"{fmax:.16f} not between 0 and 1"
        assert fmin >= 0.0 and fmin <= 1.0

        # Ensure the fraction for each image frame adds up to 1.0 for
        # all those frames completely within the grid
        for j in range(1, fraction.all()[0] - 1):
            tot = flex.sum(fraction[j : j + 1, :])
            assert abs(tot - 1.0) < 1e-7

        # Ensure the frames follow a progression through the grid. I.e,
        # check that values increase then decrease and don't jump around
        for j in range(fraction.all()[0]):
            f = fraction[j : j + 1, :]
            last = f[0]
            rev = False
            for i in range(1, len(f)):
                curr = f[1]
                if rev is False:
                    if curr < last:
                        rev = True
                else:
                    assert curr <= last
                last = curr


def test_map_frames_reverse(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )

    # Get the models
    beam = sequence.get_beam()
    detector = sequence.get_detector()
    gonio = sequence.get_goniometer()
    scan = sequence.get_scan()

    # Set the delta_divergence/mosaicity
    n_sigma = 3
    sigma_divergence = 0.060 * math.pi / 180
    mosaicity = 0.154 * math.pi / 180
    delta_divergence = n_sigma * sigma_divergence
    delta_mosaicity = n_sigma * mosaicity

    # Set the grid size
    grid_size = (4, 4, 4)

    # Create the E3 fraction object
    transform = MapFramesReverse(
        scan.get_array_range()[0],
        scan.get_oscillation(deg=False)[0],
        scan.get_oscillation(deg=False)[1],
        mosaicity,
        n_sigma,
        grid_size[2],
    )

    # Create the bounding box calculator
    calculate_bbox = BBoxCalculator3D(
        beam, detector, gonio, scan, delta_divergence, delta_mosaicity
    )

    s0 = beam.get_s0()
    m2 = gonio.get_rotation_axis()
    s0_length = matrix.col(beam.get_s0()).length()

    for i in range(100):

        # Get random x, y, z
        x = random.uniform(0, 2000)
        y = random.uniform(0, 2000)
        z = random.uniform(0, 9)

        # Get random s1, phi, panel
        s1 = matrix.col(detector[0].get_pixel_lab_coord((x, y))).normalize() * s0_length
        phi = scan.get_angle_from_array_index(z, deg=False)
        panel = 0

        # Calculate the bounding box
        bbox = calculate_bbox(s1, phi, panel)
        x1, x2 = bbox[0], bbox[1]
        y1, y2 = bbox[2], bbox[3]
        z1, z2 = bbox[4], bbox[5]
        if x1 == 0 or y1 == 0 or z1 == 0:
            continue
        if x2 == 2000 or y2 == 2000 or z2 == 9:
            continue

        # Create the XDS coordinate system
        xcs = CoordinateSystem(m2, s0, s1, phi)

        # Calculate the transform fraction
        fraction = transform(bbox[4:], phi, xcs.zeta())

        # Ensure the minimum and maximum are 0 < 1
        fmax = flex.max(fraction)
        fmin = flex.min(fraction)
        assert fmax <= 1.0 and fmax > 0.0
        assert fmin >= 0.0 and fmin <= 1.0

        # Ensure the fraction for image adds up to 1.0 for
        # all those images completely within the image
        for v3 in range(fraction.all()[0]):
            tot = flex.sum(fraction[v3 : v3 + 1, :])
            assert abs(tot - 1.0) < 1e-7

        # Ensure the frames follow a progression through the grid. I.e,
        # check that values increase then decrease and don't jump around
        for v3 in range(fraction.all()[0]):
            f = fraction[v3 : v3 + 1, :]
            last = f[0]
            rev = False
            for i in range(1, len(f)):
                curr = f[1]
                if rev is False:
                    if curr < last:
                        rev = True
                else:
                    assert curr <= last
                last = curr


def test_map_forward_reverse(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )

    # Get the models
    beam = sequence.get_beam()
    detector = sequence.get_detector()
    gonio = sequence.get_goniometer()
    scan = sequence.get_scan()

    # Set the delta_divergence/mosaicity
    n_sigma = 3
    sigma_divergence = 0.060 * math.pi / 180
    mosaicity = 0.154 * math.pi / 180
    delta_divergence = n_sigma * sigma_divergence
    delta_mosaicity = n_sigma * mosaicity

    # Set the grid size
    grid_size = (4, 4, 4)

    # Create the E3 fraction object
    transform_forward = MapFramesForward(
        scan.get_array_range()[0],
        scan.get_oscillation(deg=False)[0],
        scan.get_oscillation(deg=False)[1],
        mosaicity,
        n_sigma,
        grid_size[2],
    )

    # Create the E3 fraction object
    transform_reverse = MapFramesReverse(
        scan.get_array_range()[0],
        scan.get_oscillation(deg=False)[0],
        scan.get_oscillation(deg=False)[1],
        mosaicity,
        n_sigma,
        grid_size[2],
    )

    # Create the bounding box calculator
    calculate_bbox = BBoxCalculator3D(
        beam, detector, gonio, scan, delta_divergence, delta_mosaicity
    )

    s0 = beam.get_s0()
    m2 = gonio.get_rotation_axis()
    s0_length = matrix.col(beam.get_s0()).length()

    for i in range(100):

        # Get random x, y, z
        x = random.uniform(0, 2000)
        y = random.uniform(0, 2000)
        z = random.uniform(0, 9)

        # Get random s1, phi, panel
        s1 = matrix.col(detector[0].get_pixel_lab_coord((x, y))).normalize() * s0_length
        phi = scan.get_angle_from_array_index(z, deg=False)
        panel = 0

        # Calculate the bounding box
        bbox = calculate_bbox(s1, phi, panel)

        # Create the XDS coordinate system
        xcs = CoordinateSystem(m2, s0, s1, phi)

        # Calculate the transform fraction
        forward_fraction = transform_forward(bbox[4:], phi, xcs.zeta())

        # Calculate the transform fraction
        reverse_fraction = transform_reverse(bbox[4:], phi, xcs.zeta())

        # Check the same points are non-zero
        eps = 1e-7
        for j in range(forward_fraction.all()[0]):
            for i in range(forward_fraction.all()[1]):
                if forward_fraction[j, i] > 0.0:
                    assert reverse_fraction[i, j] > 0.0
                else:
                    assert reverse_fraction[i, j] < eps
